from collections import defaultdict
from pathlib import Path

import requests
from model_lib import Entity, parse_model

from atlas_init.cli_tf.debug_logs_test_data import ApiSpecPath, find_normalized_path
from atlas_init.cli_tf.schema import logger
from atlas_init.cli_tf.openapi import OpenapiSchema


def go_sdk_breaking_changes(repo_path: Path, go_sdk_rel_path: str = "../atlas-sdk-go") -> Path:
    rel_path = "tools/releaser/breaking_changes"
    breaking_changes_dir = repo_path / go_sdk_rel_path / rel_path
    breaking_changes_dir = breaking_changes_dir.absolute()
    assert breaking_changes_dir.exists(), f"not found breaking_changes={breaking_changes_dir}"
    return breaking_changes_dir


def api_spec_path_transformed(sdk_repo_path: Path) -> Path:
    return sdk_repo_path / "openapi/atlas-api-transformed.yaml"


class ApiSpecPaths(Entity):
    method_paths: dict[str, list[ApiSpecPath]]

    def normalize_path(self, method: str, path: str) -> str:
        if path.startswith("/api/atlas/v1.0"):
            return ""
        return find_normalized_path(path, self.method_paths[method]).path


def parse_api_spec_paths(api_spec_path: Path) -> dict[str, list[ApiSpecPath]]:
    model = parse_model(api_spec_path, t=OpenapiSchema)
    paths: dict[str, list[ApiSpecPath]] = defaultdict(list)
    for path, path_dict in model.paths.items():
        for method in path_dict:
            paths[method.upper()].append(ApiSpecPath(path=path))
    return paths


# reusing url from terraform-provider-mongodbatlas/scripts/schema-scaffold.sh
ADMIN_API_URL = "https://raw.githubusercontent.com/mongodb/atlas-sdk-go/main/openapi/atlas-api-transformed.yaml"


def admin_api_url(branch: str) -> str:
    return ADMIN_API_URL.replace("/main/", f"/{branch}/")


def download_admin_api(dest: Path, branch: str = "main") -> None:
    dest.parent.mkdir(parents=True, exist_ok=True)
    url = admin_api_url(branch)
    logger.info(f"downloading admin api to {dest} from {url}")
    response = requests.get(url, timeout=10)
    response.raise_for_status()
    dest.write_bytes(response.content)



import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from model_lib import dump
from zero_3rdparty.file_utils import copy, iter_paths_and_relative

from atlas_init.cli_helper.run import (
    run_binary_command_is_ok,
    run_command_receive_result,
)
from atlas_init.settings.config import TerraformVars, TestSuite
from atlas_init.settings.env_vars import AtlasInitSettings
from atlas_init.settings.env_vars_generated import AWSSettings, AtlasSettings

logger = logging.getLogger(__name__)


def get_tf_vars(settings: AtlasInitSettings, active_groups: list[TestSuite]) -> dict[str, Any]:  # type: ignore
    tf_vars = TerraformVars()  # type: ignore
    tf_vars = sum((group.vars for group in active_groups), start=tf_vars)
    aws_settings = AWSSettings.from_env()
    atlas_settings = AtlasSettings.from_env()
    return {
        "atlas_public_key": atlas_settings.MONGODB_ATLAS_PUBLIC_KEY,
        "atlas_private_key": atlas_settings.MONGODB_ATLAS_PRIVATE_KEY,
        "atlas_base_url": atlas_settings.MONGODB_ATLAS_BASE_URL,
        "is_mongodbgov_cloud": atlas_settings.is_mongodbgov_cloud,
        "org_id": atlas_settings.MONGODB_ATLAS_ORG_ID,
        "aws_region": aws_settings.AWS_REGION,
        "project_name": settings.project_name,
        "out_dir": settings.profile_dir,
        "extra_env_vars": settings.manual_env_vars,
        **settings.tf_vars(aws_settings.AWS_REGION),
        **tf_vars.as_configs(),
    }


class TerraformRunError(Exception):
    pass


@dataclass
class state_copier:  # noqa: N801
    state_path: Path
    tf_path: Path

    def __enter__(self):
        self.state_path.mkdir(exist_ok=True, parents=True)
        for state_path, rel_path in iter_paths_and_relative(self.state_path, "terraform.tfstate*", rglob=False):
            copy(state_path, self.tf_path / rel_path)

    def __exit__(self, *_):
        self.tf_path.mkdir(exist_ok=True, parents=True)
        for state_path, rel_path in iter_paths_and_relative(self.tf_path, "terraform.tfstate*", rglob=False):
            state_path.rename(self.state_path / rel_path)


def run_terraform(settings: AtlasInitSettings, command: str, extra_args: list[str]):
    with state_copier(settings.tf_state_path, settings.atlas_init_tf_src_path):
        _run_terraform(settings, command, extra_args)


def _run_terraform(settings: AtlasInitSettings, command: str, extra_args: list[str]):
    command_parts = [
        command,
        "-var-file",
        str(settings.tf_vars_path),
        *extra_args,
    ]
    is_ok = run_binary_command_is_ok(
        "terraform",
        " ".join(command_parts),
        env=os.environ | {"TF_DATA_DIR": settings.tf_data_dir},
        cwd=settings.atlas_init_tf_src_path,
        logger=logger,
    )
    if not is_ok:
        raise TerraformRunError


def dump_tf_vars(settings: AtlasInitSettings, tf_vars: dict[str, Any]):
    tf_vars_path = settings.tf_vars_path
    tf_vars_path.parent.mkdir(exist_ok=True, parents=True)
    tf_vars_str = dump(tf_vars, "pretty_json")
    logger.info(f"writing tf vars to {tf_vars_path}")
    logger.debug(f"tf vars:\n{tf_vars_str}")
    tf_vars_path.write_text(tf_vars_str)


def export_outputs(settings: AtlasInitSettings) -> None:
    with state_copier(settings.tf_state_path, settings.atlas_init_tf_src_path):
        result = run_command_receive_result(
            "terraform output -json",
            settings.atlas_init_tf_src_path,
            logger,
            env=os.environ | {"TF_DATA_DIR": settings.tf_data_dir},
        )
    settings.tf_outputs_path.write_text(result)



from __future__ import annotations

import fnmatch
import logging
from collections import defaultdict
from collections.abc import Iterable
from functools import total_ordering
from os import getenv
from pathlib import Path
from typing import Any

from model_lib import Entity, IgnoreFalsy
from pydantic import Field, model_validator

from atlas_init.repos.path import as_repo_alias, find_test_names, go_package_prefix, owner_project_name, package_glob

logger = logging.getLogger(__name__)


class TerraformVars(IgnoreFalsy):
    cluster_info: bool = False
    cluster_info_m10: bool = False
    stream_instance: bool = False
    use_private_link: bool = False
    use_vpc_peering: bool = False
    use_project_extra: bool = False
    use_aws_vars: bool = False
    use_aws_vpc: bool = False
    use_aws_s3: bool = False
    use_federated_vars: bool = False
    use_encryption_at_rest: bool = False

    def __add__(self, other: TerraformVars):  # type: ignore
        assert isinstance(other, TerraformVars)  # type: ignore
        kwargs = {k: v or getattr(other, k) for k, v in self}
        return type(self)(**kwargs)

    def as_configs(self) -> dict[str, Any]:
        config = {}
        if self.cluster_info or self.cluster_info_m10:
            instance_size = "M10" if self.cluster_info_m10 else "M0"
            cloud_backup = self.cluster_info_m10
            config["cluster_config"] = {
                "name": "atlas-init",
                "instance_size": instance_size,
                "database_in_url": "default",
                "cloud_backup": cloud_backup,
            }
        if self.use_private_link:
            config["use_private_link"] = True
        if self.use_vpc_peering:
            config["use_vpc_peering"] = True
        if self.use_aws_vars:
            config["use_aws_vars"] = True
        if self.use_aws_vpc:
            config["use_aws_vpc"] = True
        if self.use_aws_s3:
            config["use_aws_s3"] = True
        if self.use_project_extra:
            config["use_project_extra"] = True
        if self.use_federated_vars:
            config["use_federated_vars"] = True
        if self.use_encryption_at_rest:
            config["use_encryption_at_rest"] = True
        if self.stream_instance:
            # hack until backend bug with stream instance is fixed
            config["stream_instance_config"] = {"name": getenv("ATLAS_STREAM_INSTANCE_NAME", "atlas-init")}
        return config


class PyHook(Entity):
    name: str
    locate: str


@total_ordering
class TestSuite(IgnoreFalsy):
    __test__ = False

    name: str
    sequential_tests: bool = False
    repo_go_packages: dict[str, list[str]] = Field(default_factory=dict)
    vars: TerraformVars = Field(default_factory=TerraformVars)  # type: ignore
    post_apply_hooks: list[PyHook] = Field(default_factory=list)

    def __lt__(self, other) -> bool:
        if not isinstance(other, TestSuite):  # type: ignore
            raise TypeError
        return self.name < other.name

    def package_url_tests(self, repo_path: Path, prefix: str = "") -> dict[str, dict[str, Path]]:
        alias = as_repo_alias(repo_path)
        packages = self.repo_go_packages.get(alias, [])
        names = defaultdict(dict)
        for package in packages:
            pkg_name = f"{go_package_prefix(repo_path)}/{package}"
            for go_file in repo_path.glob(f"{package}/*.go"):
                for name in find_test_names(go_file, prefix):
                    names[pkg_name][name] = go_file.parent
        return names

    def is_active(self, repo_alias: str, change_paths: Iterable[str]) -> bool:
        """changes paths should be relative to the repo"""
        globs = [package_glob(pkg) for pkg in self.repo_go_packages.get(repo_alias, [])]
        return any(any(fnmatch.fnmatch(path, glob) for glob in globs) for path in change_paths)

    def cwd_is_repo_go_pkg(self, cwd: Path, repo_alias: str) -> bool:
        alias_packages = self.repo_go_packages[repo_alias]
        for pkg_path in alias_packages:
            if str(cwd).endswith(pkg_path):
                return True
        logger.warning(f"no go package found for repo {repo_alias} in {cwd}")
        return False


class RepoAliasNotFoundError(ValueError):
    def __init__(self, name: str) -> None:
        self.name = name
        super().__init__(name)


class AtlasInitConfig(Entity):
    repo_aliases: dict[str, str] = Field(default_factory=dict)
    test_suites: list[TestSuite] = Field(default_factory=list)  # type: ignore

    def repo_alias(self, repo_url_path: str) -> str:
        alias = self.repo_aliases.get(repo_url_path)
        if alias is None:
            raise RepoAliasNotFoundError(repo_url_path)
        return alias

    def go_package_prefix(self, alias: str) -> str:
        for url_path, i_alias in self.repo_aliases.items():
            if alias == i_alias:
                return f"github.com/{url_path}"
        raise ValueError(f"alias not found: {alias}")

    def active_test_suites(
        self,
        alias: str | None,
        change_paths: Iterable[str],
        forced_test_suites: list[str],
    ) -> list[TestSuite]:  # type: ignore
        forced_suites = set(forced_test_suites)
        if forced_test_suites:
            logger.warning(f"using forced test suites: {forced_test_suites}")
        return [
            suit
            for suit in self.test_suites
            if suit.name in forced_suites or (alias and suit.is_active(alias, change_paths))
        ]

    @model_validator(mode="after")
    def ensure_all_repo_aliases_are_found(self):
        missing_aliases = set()
        aliases = set(self.repo_aliases.values())
        for group in self.test_suites:
            if more_missing := (group.repo_go_packages.keys() - aliases):
                logger.warning(f"repo aliases not found for group={group.name}: {more_missing}")
                missing_aliases |= more_missing
        if missing_aliases:
            raise ValueError(f"repo aliases not found: {missing_aliases}")
        return self


def active_suites(
    config: AtlasInitConfig,
    repo_path: Path,
    cwd_rel_path: str,
    forced_test_suites: list[str],
) -> list[TestSuite]:  # type: ignore
    repo_url_path = owner_project_name(repo_path)
    try:
        repo_alias = config.repo_alias(repo_url_path)
    except RepoAliasNotFoundError:
        if forced_test_suites:
            # still want to use the forced test suites
            repo_alias = None
        else:
            raise
    logger.info(
        f"repo_alias={repo_alias}, repo_path={repo_path}, repo_url_path={repo_url_path}, cwd_rel_path={cwd_rel_path}"
    )
    change_paths = [cwd_rel_path]
    active_suites = config.active_test_suites(repo_alias, change_paths, forced_test_suites)
    logger.info(f"active_suites: {[s.name for s in active_suites]}")
    return active_suites



from __future__ import annotations

import datetime
import logging
import re
from collections.abc import Iterable
from pathlib import Path
from queue import Queue
from typing import ClassVar, NamedTuple

from model_lib import Entity, dump
from pydantic import Field

from atlas_init.cli_tf.schema_v2 import (
    NewAttribute,
    SchemaAttribute,
    SchemaResource,
    SchemaV2,
    SkipAttribute,
    parse_model,
)

logger = logging.getLogger(__name__)


def api_spec_text_changes(schema: SchemaV2, api_spec_parsed: OpenapiSchema) -> OpenapiSchema:
    openapi_changes = schema.openapi_changes
    schema_to_update = api_spec_parsed.components.get("schemas", {})
    original_schema = {**api_spec_parsed.components.get("schemas", {})}  # copy used for iteration
    assert isinstance(original_schema, dict), "Expected a dict @ components.schemas"
    for prefix in openapi_changes.schema_prefix_removal:
        for name, value in original_schema.items():
            if name.startswith(prefix):
                schema_to_update.pop(name)
                name_no_prefix = name.removeprefix(prefix)
                assert name_no_prefix not in schema_to_update, (
                    f"removed {prefix} from {name} in schema but {name_no_prefix} already exists"
                )
                schema_to_update[name_no_prefix] = value
    openapi_yaml = dump(api_spec_parsed, "yaml")
    for prefix in openapi_changes.schema_prefix_removal:
        pattern = re.compile(rf"{OpenapiSchema.SCHEMAS_PREFIX}(?P<prefix>{prefix})(?P<name>\w+)")
        openapi_yaml = pattern.sub(rf"{OpenapiSchema.SCHEMAS_PREFIX}\g<name>", openapi_yaml)
    return parse_model(openapi_yaml, t=OpenapiSchema, format="yaml")


def parse_openapi_schema_after_modifications(schema: SchemaV2, api_spec_path: Path) -> OpenapiSchema:
    original = parse_model(api_spec_path, t=OpenapiSchema)
    return api_spec_text_changes(schema, original)


class PathMethodCode(NamedTuple):
    path: str
    method: str
    code: str


def extract_api_version_content_header(header: str) -> str | None:
    """
    Extracts the API version from the content header.
    The header should be in the format 'application/vnd.atlas.v1+json'.
    """
    match = re.match(r"application/vnd\.atlas\.v?(?P<version>[\d-]+)\+json", header)
    return match.group("version") if match else None


class OpenapiSchema(Entity):
    PARAMETERS_PREFIX: ClassVar[str] = "#/components/parameters/"
    SCHEMAS_PREFIX: ClassVar[str] = "#/components/schemas/"

    openapi: str
    info: dict
    paths: dict
    components: dict
    tags: list = Field(default_factory=list)

    def create_method(self, path: str) -> dict | None:
        return self.paths.get(path, {}).get("post")

    def get_method(self, path: str) -> dict | None:
        return self.paths.get(path, {}).get("get")

    def delete_method(self, path: str) -> dict | None:
        return self.paths.get(path, {}).get("delete")

    def patch_method(self, path: str) -> dict | None:
        return self.paths.get(path, {}).get("patch")

    def put_method(self, path: str) -> dict | None:
        return self.paths.get(path, {}).get("patch")

    def methods_with_name(self, path: str) -> Iterable[tuple[str, dict]]:
        for method_name in ["post", "get", "delete", "patch", "put"]:
            if method := self.paths.get(path, {}).get(method_name):
                yield method_name, method

    def methods(self, path: str) -> Iterable[dict]:
        yield from (method for _, method in self.methods_with_name(path))

    def method_refs(self, path: str) -> Iterable[str]:
        for method in self.methods(path):
            if method:
                yield from self.method_request_body_ref(method)
                yield from self.method_response_ref(method)

    def parameter_refs(self, path: str) -> Iterable[str]:
        for method in self.methods(path):
            parameters = method.get("parameters", [])
            for param in parameters:
                if param_ref := param.get("$ref"):
                    yield param_ref

    def parameter(self, ref: str) -> dict:
        assert ref.startswith(OpenapiSchema.PARAMETERS_PREFIX)
        parameter_name = ref.split("/")[-1]
        param_dict = self.components["parameters"][parameter_name]
        assert isinstance(param_dict, dict), f"Expected a dict @ {ref}"
        return param_dict

    def schema_properties(self, ref: str) -> Iterable[dict]:
        assert ref.startswith(OpenapiSchema.SCHEMAS_PREFIX)
        schema_name = ref.split("/")[-1]
        schema_dict = self.components["schemas"][schema_name]
        assert isinstance(schema_dict, dict), f"Expected a dict @ {ref}"
        properties = schema_dict.get("properties", {})
        assert isinstance(properties, dict), f"Expected a dict @ {ref}.properties"
        for name, prop in properties.items():
            assert isinstance(prop, dict), f"Expected a dict @ {ref}.properties.{name}"
            prop["name"] = name
            yield prop

    def method_request_body_ref(self, method: dict) -> Iterable[str]:
        request_body = method.get("requestBody", {})
        yield from self._unpack_schema_ref(request_body)

    def method_response_ref(self, method: dict) -> Iterable[str]:
        responses = method.get("responses", {})
        ok_response = responses.get("200", {})
        yield from self._unpack_schema_ref(ok_response)

    def _unpack_schema_ref(self, response: dict) -> Iterable[str]:
        content = {**response.get("content", {})}  # avoid side effects
        if not content:
            return None
        while content:
            key, value = content.popitem()
            if not isinstance(key, str) or not key.endswith("json"):
                continue
            if ref := value.get("schema", {}).get("$ref"):
                yield ref

    def _unpack_schema_versions(self, response: dict) -> list[datetime.date]:
        content: dict[str, dict] = {**response.get("content", {})}
        versions = []
        while content:
            key, value = content.popitem()
            if not isinstance(value, dict) or not key.endswith("json"):
                continue
            if version := value.get("x-xgen-version"):
                versions.append(version)
                continue
            if version := extract_api_version_content_header(key):
                versions.append(version)
        return versions

    def path_method_api_versions(self) -> Iterable[tuple[PathMethodCode, list[datetime.date]]]:
        for path, methods in self.paths.items():
            for method_name, method_dict in methods.items():
                if not isinstance(method_dict, dict):
                    continue
                responses = method_dict.get("responses", {})
                for code, response_dict in responses.items():
                    if api_versions := self._unpack_schema_versions(response_dict):
                        key = PathMethodCode(path, method_name, code)
                        yield key, api_versions

    def schema_ref_component(self, ref: str, attributes_skip: set[str]) -> SchemaResource:
        schemas = self.components.get("schemas", {})
        assert isinstance(schemas, dict), "Expected a dict @ components.schemas"
        schema = schemas.get(ref.split("/")[-1])
        assert isinstance(schema, dict), f"Expected a dict @ components.schemas.{ref}"
        return self._as_schema_resource(attributes_skip, schema, ref)

    def schema_ref_components(self, attributes_skip: set[str]) -> Iterable[SchemaResource]:
        schemas = self.components.get("schemas", {})
        assert isinstance(schemas, dict), "Expected a dict @ components.schemas"
        for name, schema in schemas.items():
            ref = f"#/components/schemas/{name}"
            yield self._as_schema_resource(attributes_skip, schema, ref)

    def _as_schema_resource(self, attributes_skip: set[str], schema: dict, ref: str):
        schema_resource = SchemaResource(
            name=ref,
            description=schema.get("description", ""),
            attributes_skip=attributes_skip,
            discriminator=schema.get("discriminator"),
            one_of=schema.get("oneOf", []),
            all_of=schema.get("allOf", []),
        )
        required_names = schema.get("required", [])
        for prop in self.schema_properties(ref):
            if attr := parse_api_spec_param(self, prop, schema_resource):
                attr.is_required = prop["name"] in required_names
                schema_resource.attributes[attr.name] = attr
        return schema_resource

    def add_schema_ref(self, ref: str, ref_value: dict) -> None:
        if ref.startswith(self.PARAMETERS_PREFIX):
            prefix = self.PARAMETERS_PREFIX
            parent_dict = self.components["parameters"]
        elif ref.startswith(self.SCHEMAS_PREFIX):
            prefix = self.SCHEMAS_PREFIX
            parent_dict = self.components["schemas"]
            ref_value.pop("name", None)
            if properties := ref_value.get("properties"):
                properties_no_name = {
                    k: {nested_k: nested_v for nested_k, nested_v in v.items() if nested_k != "name"}
                    for k, v in properties.items()
                }
                if ref.removeprefix(prefix).endswith("DBRoleToExecute"):
                    logger.warning(f"debug me: {properties_no_name}")
                ref_value["properties"] = properties_no_name

        else:
            err_msg = f"Unknown schema_ref {ref}"
            raise ValueError(err_msg)
        parent_dict[ref.removeprefix(prefix)] = ref_value

    def resolve_ref(self, ref: str) -> dict:
        if ref.startswith(self.PARAMETERS_PREFIX):
            return self.parameter(ref)
        if ref.startswith(self.SCHEMAS_PREFIX):
            return self.components["schemas"][ref.split("/")[-1]]
        err_msg = f"Unknown ref {ref}"
        raise ValueError(err_msg)


def parse_api_spec_param(api_spec: OpenapiSchema, param: dict, resource: SchemaResource) -> SchemaAttribute | None:
    match param:
        case {"$ref": ref} if ref.startswith(OpenapiSchema.PARAMETERS_PREFIX):
            param_root = api_spec.parameter(ref)
            found_attribute = parse_api_spec_param(api_spec, param_root, resource)
            if found_attribute:
                found_attribute.parameter_ref = ref
            return found_attribute
        case {"$ref": ref, "name": name} if ref.startswith(OpenapiSchema.SCHEMAS_PREFIX):
            # nested attribute
            attribute = SchemaAttribute(
                additional_properties=param.get("additionalProperties", {}),
                type="object",
                name=name,
                schema_ref=ref,
            )
        case {"type": "array", "items": {"$ref": ref}, "name": name}:
            attribute = SchemaAttribute(
                additional_properties=param.get("additionalProperties", {}),
                type="array",
                name=name,
                schema_ref=ref,
                description=param.get("description", ""),
                is_computed=param.get("readOnly", False),
                is_required=param.get("required", False),
            )
        case {"name": name, "schema": schema}:
            attribute = SchemaAttribute(
                additional_properties=param.get("additionalProperties", {}),
                type=schema["type"],
                name=name,
                description=param.get("description", ""),
                is_computed=schema.get("readOnly", False),
                is_required=param.get("required", False),
            )
        case {"name": name, "type": type_}:
            attribute = SchemaAttribute(
                type=type_,
                name=name,
                description=param.get("description", ""),
                is_computed=param.get("readOnly", False),
                is_required=param.get("required", False),
                additional_properties=param.get("additionalProperties", {}),
            )
        case _:
            raise NotImplementedError
    try:
        existing = resource.lookup_attribute(attribute.name)
        logger.info(f"Merging attribute {attribute.name} into {existing.name}")
        attribute = existing.merge(attribute)
    except NewAttribute:
        logger.info("Adding new attribute %s to %s", attribute.name, resource.name)
    except SkipAttribute:
        return None
    resource.attributes[attribute.name] = attribute
    return attribute


def add_api_spec_info(schema: SchemaV2, api_spec_path: Path, *, minimal_refs: bool = False) -> None:
    api_spec = parse_openapi_schema_after_modifications(schema, api_spec_path)
    for resource in schema.resources.values():
        for path in resource.paths:
            create_method = api_spec.create_method(path)
            if not create_method:
                continue
            for param in create_method.get("parameters", []):
                parse_api_spec_param(api_spec, param, resource)
            for req_ref in api_spec.method_request_body_ref(create_method):
                for property_dict in api_spec.schema_properties(req_ref):
                    parse_api_spec_param(api_spec, property_dict, resource)
        for path in resource.paths:
            read_method = api_spec.get_method(path)
            if not read_method:
                continue
            for param in read_method.get("parameters", []):
                parse_api_spec_param(api_spec, param, resource)
            for response_ref in api_spec.method_response_ref(read_method):
                for property_dict in api_spec.schema_properties(response_ref):
                    parse_api_spec_param(api_spec, property_dict, resource)
    if minimal_refs:
        minimal_ref_resources(schema, api_spec)
    else:
        for resource in api_spec.schema_ref_components(schema.attributes_skip):
            schema.ref_resources[resource.name] = resource


def minimal_ref_resources(schema: SchemaV2, api_spec: OpenapiSchema) -> None:
    include_refs = Queue()
    seen_refs = set()
    for resource in schema.resources.values():
        for attribute in resource.attributes.values():
            if attribute.schema_ref:
                include_refs.put(attribute.schema_ref)
    while not include_refs.empty():
        ref = include_refs.get()
        logger.info(f"Adding ref {ref}")
        seen_refs.add(ref)
        ref_resource = api_spec.schema_ref_component(ref, schema.attributes_skip)
        schema.ref_resources[ref_resource.name] = ref_resource
        for attribute in ref_resource.attributes.values():
            if attribute.schema_ref and attribute.schema_ref not in seen_refs:
                include_refs.put(attribute.schema_ref)


def minimal_api_spec(schema: SchemaV2, original_api_spec_path: Path) -> OpenapiSchema:
    schema.reset_attributes_skip()
    full_spec = parse_openapi_schema_after_modifications(schema, original_api_spec_path)
    add_api_spec_info(schema, original_api_spec_path, minimal_refs=True)
    minimal_spec = OpenapiSchema(
        openapi=full_spec.openapi,
        info=full_spec.info,
        paths={},
        components={"schemas": {}, "parameters": {}},
    )
    include_refs = Queue()
    seen_refs = set()

    def add_from_resource(resource: SchemaResource) -> None:
        for path in resource.paths:
            minimal_spec.paths[path] = full_spec.paths[path]
            for ref in full_spec.method_refs(path):
                minimal_spec.add_schema_ref(ref, full_spec.resolve_ref(ref))
                include_refs.put(ref)
        for attribute in resource.attributes.values():
            if attribute.schema_ref:
                minimal_spec.add_schema_ref(attribute.schema_ref, full_spec.resolve_ref(attribute.schema_ref))
                include_refs.put(attribute.schema_ref)
            if attribute.parameter_ref:
                minimal_spec.add_schema_ref(
                    attribute.parameter_ref,
                    full_spec.resolve_ref(attribute.parameter_ref),
                )

    for resource in schema.resources.values():
        add_from_resource(resource)
    while not include_refs.empty():
        ref = include_refs.get()
        if ref in seen_refs:
            continue
        seen_refs.add(ref)
        ref_resource = full_spec.schema_ref_component(ref, schema.attributes_skip)
        add_from_resource(ref_resource)
    return minimal_spec



import logging
from pathlib import Path

import requests
import typer
from atlas_init.typer_app import app_command
from git import Repo

logger = logging.getLogger(__name__)


@app_command()
def mms_released(
    mms_repo: Path = typer.Option(..., "-r", "--mms-repo", help="the path to the mms repo"),
    commit_shas: list[str] = typer.Option(
        ..., "-c", "--commit-shas", help="the commit shas to check for release, can be set multiple times"
    ),
    sha_url: str = typer.Option(
        "https://cloud.mongodb.com/api/private/unauth/version",
        "-u",
        "--url",
        help="the url to get the current sha from",
    ),
):
    assert mms_repo.exists(), f"mms repo not found @ {mms_repo}"
    git_repo = Repo(mms_repo)
    git_repo.git.fetch("origin")
    for sha in commit_shas:
        assert git_repo.commit(sha), f"commit {sha} not found in {mms_repo}"
    current_sha_response = requests.get(sha_url, timeout=10)
    current_sha_response.raise_for_status()
    current_sha = current_sha_response.text.strip()
    assert current_sha, f"unable to get current sha from {current_sha_response.url}"
    logger.info(f"current sha of prod: {current_sha}")
    assert git_repo.commit(current_sha)
    remaining_shas = set(commit_shas)
    for commit in git_repo.iter_commits(rev=current_sha):
        commit_sha = commit.hexsha
        if commit_sha in commit_shas:
            commit_message = commit.message.rstrip("\n")  # type: ignore
            logger.info(f"found commit {commit_sha} with message {commit_message}")
            remaining_shas.remove(commit_sha)
        if not remaining_shas:
            logger.info("all commits found ✅")
            return
    logger.info(f"remaining shas: {','.join(sorted(remaining_shas))} ❌")



from __future__ import annotations

import logging
import os
import signal
import subprocess  # nosec
import sys
import threading
import time
from collections import deque
from collections.abc import Callable
from concurrent.futures import Future, ThreadPoolExecutor
from dataclasses import dataclass, field
from logging import Logger
from pathlib import Path
from time import monotonic

from atlas_init.cli_helper.run import LOG_CMD_PREFIX, find_binary_on_path

default_logger = logging.getLogger(__name__)


class ResultInProgressError(Exception):
    pass


class ResultDoneError(Exception):
    pass


class LogTextNotFoundError(Exception):
    def __init__(self, store: ResultStore) -> None:
        self.store = store
        super().__init__(store)


@dataclass
class WaitOnText:
    line: str
    timeout: float


@dataclass
class ResultStore:
    wait_condition: WaitOnText | None = None
    _recent_lines: deque = field(default_factory=lambda: deque(maxlen=1000))

    result: list[str] = field(default_factory=list)
    exit_code: int | None = None
    _aborted: bool = False
    _terminated: bool = False
    _killed: bool = False

    @property
    def result_str(self) -> str:
        return "".join(self.result)

    @property
    def is_ok(self) -> bool:
        if self.in_progress():
            raise ResultInProgressError
        return self.exit_code == 0

    def _add_line(self, line: str) -> None:
        self._recent_lines.append(line)
        self.result.append(line)

    def unexpected_error(self) -> bool:
        if self.in_progress():
            raise ResultInProgressError
        return self.exit_code != 0 and not self._aborted

    def force_stopped(self) -> bool:
        if self.in_progress():
            raise ResultInProgressError
        return self._killed or self._terminated

    def in_progress(self) -> bool:
        return self.exit_code is None

    def wait(self) -> None:
        condition = self.wait_condition
        if not condition:
            return
        timeout = condition.timeout
        start = monotonic()
        while monotonic() - start < timeout:
            if not self.in_progress():
                raise LogTextNotFoundError(self)
            while self._recent_lines:
                line = self._recent_lines.popleft()
                if condition.line in line:
                    return
            time.sleep(0.1)
        raise LogTextNotFoundError(self)

    def _abort(self) -> None:
        self._aborted = True

    def _terminate(self) -> None:
        self._terminated = True

    def _kill(self) -> None:
        self._killed = True


class RunManager:
    def __init__(
        self,
        worker_count: int = 100,
        signal_int_timeout_s: float = 0.2,
        signal_term_timeout_s: float = 0.2,
        signal_kill_timeout_s: float = 0.2,
        *,
        dry_run: bool = False,
    ):
        """
        Args:
            worker_count: the number of workers to run in parallel
            terminate_read_timeout: the time to wait after terminating a process before closing the output
        """
        self.processes: dict[int, subprocess.Popen] = {}
        self.results: dict[int, ResultStore] = {}
        self.lock = threading.RLock()
        self.pool = ThreadPoolExecutor(max_workers=worker_count)
        self.signal_int_timeout_s = signal_int_timeout_s
        self.signal_term_timeout_s = signal_term_timeout_s
        self.signal_kill_timeout_s = signal_kill_timeout_s
        self.dry_run = dry_run

    def set_timeouts(self, timeout: float):
        self.signal_int_timeout_s = timeout
        self.signal_term_timeout_s = timeout
        self.signal_kill_timeout_s = timeout

    def __enter__(self):
        self.pool.__enter__()
        return self

    def run_process_wait_on_log(
        self,
        command: str,
        cwd: Path,
        logger: Logger | None,
        env: dict | None = None,
        result_store: ResultStore | None = None,
        *,
        line_in_log: str,
        timeout: float,
        binary: str = "",
    ) -> Future[ResultStore]:
        command = self._resolve_command(binary, command, logger)
        store = result_store or ResultStore()
        store.wait_condition = WaitOnText(line=line_in_log, timeout=timeout)
        future = self.pool.submit(self._run, command, cwd, logger, env, store)
        if not self.dry_run:
            store.wait()
        return future

    def run_process(
        self,
        command: str,
        cwd: Path,
        logger: Logger | None,
        env: dict | None = None,
        result_store: ResultStore | None = None,
        *,
        binary: str = "",
    ) -> Future[ResultStore]:
        command = self._resolve_command(binary, command, logger)
        return self.pool.submit(self._run, command, cwd, logger, env, result_store)

    def _resolve_command(self, binary: str, command: str, logger: Logger | None):
        if binary:
            binary_path = find_binary_on_path(binary, logger or default_logger, allow_missing=self.dry_run)
            command = f"{binary_path} {command}"
        return command

    def _run(
        self,
        command: str,
        cwd: Path,
        logger: Logger | None,
        env: dict | None = None,
        result: ResultStore | None = None,
    ) -> ResultStore:
        result = result or ResultStore()
        logger = logger or default_logger

        def read_output(process: subprocess.Popen):
            for line in process.stdout:  # type: ignore
                result._add_line(line)

        sys_stderr = sys.stderr

        def read_stderr(process: subprocess.Popen):
            for line in process.stderr:  # type: ignore
                sys_stderr.write(line)
                result._add_line(line)

        logger.info(f"{LOG_CMD_PREFIX}{command}' from '{cwd}'")
        if self.dry_run:
            result.exit_code = 0
            result.result.append(f"DRY RUN: {command}")
            return result
        with subprocess.Popen(
            command,
            cwd=cwd,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=sys.stdin,
            start_new_session=True,
            shell=True,  # noqa: S602 # We control the calls to this function and don't suspect any shell injection #nosec
            bufsize=0,
            text=True,  # This makes it return strings instead of bytes
        ) as process:
            with self.lock:
                self.processes[threading.get_ident()] = process
                self.results[threading.get_ident()] = result
            read_future_out = self.pool.submit(read_output, process)
            read_future_err = self.pool.submit(read_stderr, process)
            try:
                process.wait()
            except Exception:
                logger.exception(f"failed to run command: {command}")
            finally:
                for std_name, future in zip(["stdout", "stderr"], [read_future_out, read_future_err], strict=False):
                    try:
                        future.result(1)
                    except BaseException:
                        logger.exception(f"failed to read output ({std_name}) for command: {command}")
                with self.lock:
                    del self.processes[threading.get_ident()]
                    del self.results[threading.get_ident()]
        result.exit_code = process.returncode
        if result.unexpected_error():
            logger.error(f"command failed '{command}', error code: {result.exit_code}")
        if result.force_stopped():
            logger.error(f"command killed '{command}'")
        return result

    def __exit__(self, *_):
        self.pool.shutdown(wait=False, cancel_futures=True)  # wait happens in __exit__, avoid new futures starting
        self.terminate_all()
        self.pool.__exit__(None, None, None)

    def terminate_all(self):
        self._send_signal_to_all(signal.SIGINT, ResultStore._abort)
        self.wait_for_processes_ok(self.signal_int_timeout_s)
        self._send_signal_to_all(signal.SIGTERM, ResultStore._terminate)
        self.wait_for_processes_ok(self.signal_term_timeout_s)
        self._send_signal_to_all(signal.SIGKILL, ResultStore._kill)
        self.wait_for_processes_ok(self.signal_kill_timeout_s)

    def _send_signal_to_all(self, signal_type: signal.Signals, result_call: Callable[[ResultStore], None]):
        with self.lock:
            for pid, process in self.processes.items():
                result_call(self.results[pid])
                gpid = os.getpgid(process.pid)
                os.killpg(gpid, signal_type)

    def wait_for_processes_ok(self, timeout: float):
        start = monotonic()
        if not self.processes:
            return True
        while monotonic() - start < timeout:
            with self.lock:
                if not any(result.in_progress() for result in self.results.values()):
                    return True
            time.sleep(0.1)
        return False



import logging
import os
from collections.abc import Callable
from pathlib import Path

import dotenv
from appdirs import user_data_dir
from zero_3rdparty.file_utils import ensure_parents_write_text

from atlas_init import running_in_repo

logger = logging.getLogger(__name__)
"""WARNING these variables should only be used through the AtlasInitSettings, not directly"""
if running_in_repo():
    ROOT_PATH = Path(__file__).parent.parent.parent  # atlas_init REPO_PATH
else:
    ROOT_PATH = Path(__file__).parent.parent  # site package install directory
    _default_profiles_path = os.environ.get("ATLAS_INIT_PROFILES_PATH")
    if not _default_profiles_path:
        _default_profiles_path = Path(user_data_dir("atlas_init")) / "profiles"
        warning_msg = f"os.environ['ATLAS_INIT_PROFILES_PATH'] is not set using default: {_default_profiles_path}"
        logger.warning(warning_msg)
DEFAULT_TF_SRC_PATH = ROOT_PATH / "tf"
DEFAULT_ATLAS_INIT_CONFIG_PATH = ROOT_PATH / "atlas_init.yaml"
DEFAULT_ATLAS_INIT_SCHEMA_CONFIG_PATH = ROOT_PATH / "terraform.yaml"


def load_dotenv(env_path: Path) -> dict[str, str]:
    return {k: v for k, v in dotenv.dotenv_values(env_path).items() if v}


def dump_vscode_dotenv(generated_path: Path, vscode_env_path: Path, **extras: str) -> None:
    vscode_env_vars = load_dotenv(generated_path)
    vscode_env_vars.pop("TF_CLI_CONFIG_FILE", None)  # migration tests will use local provider instead of online
    vscode_env_vars.update(extras)
    dump_dotenv(vscode_env_path, vscode_env_vars)


def dump_dotenv(path: Path, env_vars: dict[str, str]):
    ensure_parents_write_text(path, "")
    for k, v in env_vars.items():
        dotenv.set_key(path, k, v)


def current_dir():
    return Path(os.path.curdir).absolute()


def default_factory_cwd(rel_path: str) -> Callable[[], Path]:
    def default_factory():
        return current_dir() / rel_path

    return default_factory


def repo_path_rel_path() -> tuple[Path, str]:
    cwd = current_dir()
    rel_path = []
    for path in [cwd, *cwd.parents]:
        if (path / ".git").exists():
            return path, "/".join(reversed(rel_path))
        rel_path.append(path.name)
    msg = "no repo path found from cwd"
    raise CwdIsNoRepoPathError(msg)


class CwdIsNoRepoPathError(ValueError):
    pass



import typer


def default_skippped_directories() -> list[str]:
    return [
        "prometheus-and-teams",  #  Provider registry.terraform.io/hashicorp/template v2.2.0 does not have a package available for your current platform, darwin_arm64.
    ]


REPO_PATH_ARG = typer.Argument(help, help="Path to the mongodbatlas-terraform-provider repository")
SKIP_EXAMPLES_DIRS_OPTION = typer.Option(
    ...,
    "--skip-examples",
    help="Skip example directories with these names",
    default_factory=default_skippped_directories,
    show_default=True,
)



from pathlib import Path
from model_lib import StaticSettings


class TfDepSettings(StaticSettings):
    @property
    def atlas_graph_path(self) -> Path:
        return self.static_root / "atlas_graph.yaml"

    @property
    def vars_file_path(self) -> Path:
        return self.static_root / "tf_vars.yaml"

    @property
    def vars_external_file_path(self) -> Path:
        return self.static_root / "tf_vars_external.yaml"

    @property
    def resource_types_file_path(self) -> Path:
        return self.static_root / "tf_resource_types.yaml"

    @property
    def resource_types_external_file_path(self) -> Path:
        return self.static_root / "tf_resource_types_external.yaml"

    @property
    def schema_resource_types_path(self) -> Path:
        return self.static_root / "tf_schema_resource_types.yaml"

    @property
    def schema_resource_types_deprecated_path(self) -> Path:
        return self.static_root / "tf_schema_resource_types_deprecated.yaml"

    @property
    def api_calls_path(self) -> Path:
        return self.static_root / "tf_api_calls.yaml"

    def pagination_output_path(self, query_string: str) -> Path:
        return self.static_root / "pagination_output" / f"query_is_{query_string or 'empty'}.md"



from __future__ import annotations
import logging
import re
from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass
from enum import StrEnum
from functools import lru_cache
from pathlib import Path
from typing import NamedTuple

from git import Repo as _GitRepo

from atlas_init.settings.path import current_dir, repo_path_rel_path

logger = logging.getLogger(__name__)
GH_OWNER_TERRAFORM_PROVIDER_MONGODBATLAS = "mongodb/terraform-provider-mongodbatlas"
GH_OWNER_MONGODBATLAS_CLOUDFORMATION_RESOURCES = "mongodb/mongodbatlas-cloudformation-resources"
_KNOWN_OWNER_PROJECTS = {
    GH_OWNER_MONGODBATLAS_CLOUDFORMATION_RESOURCES,
    GH_OWNER_TERRAFORM_PROVIDER_MONGODBATLAS,
}


def package_glob(package_path: str) -> str:
    return f"{package_path}/*.go"


def go_package_prefix(repo_path: Path) -> str:
    owner_project = owner_project_name(repo_path)
    return f"github.com/{owner_project}"


def _owner_project_name(repo_path: Path) -> str:
    owner_project = owner_project_name(repo_path)
    if owner_project not in _KNOWN_OWNER_PROJECTS:
        raise ValueError(f"unknown repo owner @ {repo_path}")
    return owner_project


_resource_roots = {
    GH_OWNER_TERRAFORM_PROVIDER_MONGODBATLAS: lambda p: p / "internal/service",
    GH_OWNER_MONGODBATLAS_CLOUDFORMATION_RESOURCES: lambda p: p / "cfn-resources",
}


def _default_is_resource(p: Path) -> bool:
    return "internal/service" in str(p)


_resource_is_resource: dict[str, Callable[[Path], bool]] = {
    GH_OWNER_MONGODBATLAS_CLOUDFORMATION_RESOURCES: lambda p: (p / "cmd/main.go").exists(),
    GH_OWNER_TERRAFORM_PROVIDER_MONGODBATLAS: _default_is_resource,
}


def resource_root(repo_path: Path) -> Path:
    owner_project = owner_project_name(repo_path)
    return _resource_roots[owner_project](repo_path)


def is_resource_call(repo_path: Path) -> Callable[[Path], bool]:
    owner_project = owner_project_name(repo_path)
    return _resource_is_resource[owner_project]


def resource_dir(repo_path: Path, full_path: Path) -> Path:
    if dir_name := resource_name(repo_path, full_path):
        return resource_root(repo_path) / dir_name
    else:
        raise ValueError(f"no resource name for {full_path}")


class Repo(StrEnum):
    CFN = "cfn"
    TF = "tf"


def as_repo_alias(path: Path) -> Repo:
    owner = owner_project_name(path)
    return _owner_lookup(owner)


_owner_repos = {
    GH_OWNER_TERRAFORM_PROVIDER_MONGODBATLAS: Repo.TF,
    GH_OWNER_MONGODBATLAS_CLOUDFORMATION_RESOURCES: Repo.CFN,
}


def _owner_lookup(owner: str) -> Repo:
    if repo := _owner_repos.get(owner):
        return repo
    raise ValueError(f"unknown repo: {owner}")


@lru_cache
def owner_project_name(repo_path: Path) -> str:
    repo = _GitRepo(repo_path)
    remote = repo.remotes[0]
    repo_url = next(iter(remote.urls))
    repo_url = repo_url.removesuffix(".git")
    *_, owner, project_name = repo_url.split("/")
    if ":" in owner:
        owner = owner.rsplit(":")[1]
    return f"{owner}/{project_name}"


def current_repo() -> Repo:
    repo_path = _repo_path()
    owner = owner_project_name(repo_path)
    return _owner_lookup(owner)


class ResourcePaths(NamedTuple):
    repo_path: Path
    resource_path: Path
    resource_name: str


def resource_name(repo_path: Path, full_path: Path) -> str:
    root = resource_root(repo_path)
    is_resource = is_resource_call(repo_path)
    if not root.exists():
        raise ValueError(f"no resource root found for {repo_path}")
    return next(
        (parent.name for parent in [full_path, *full_path.parents] if parent.parent == root and is_resource(parent)),
        "",
    )


def find_paths(assert_repo: Repo | None = None) -> ResourcePaths:
    repo_path = current_repo_path(assert_repo)
    cwd = current_dir()
    resource_path = resource_dir(repo_path, cwd)
    r_name = resource_name(repo_path, cwd)
    return ResourcePaths(repo_path, resource_path, r_name)


def _assert_repo(expected: Repo | None = None):
    if expected and current_repo() != expected:
        raise ValueError(f"wrong repo, expected {expected} and got {current_repo()}")


def current_repo_path(assert_repo: Repo | None = None) -> Path:
    _assert_repo(assert_repo)
    return _repo_path()


def _repo_path() -> Path:
    repo_path, _ = repo_path_rel_path()
    return repo_path


def find_go_mod_dir(repo_path: Path):
    for go_mod in repo_path.rglob("go.mod"):
        # tf is at root level
        # cfn is at {root}/cfn-resources
        if repo_path in {go_mod.parent, go_mod.parent.parent}:
            return go_mod.parent
    msg = "go.mod not found or more than 1 level deep"
    raise ValueError(msg)


def find_test_names(file: Path, prefix: str = "Test") -> list[str]:
    test_names = []
    with file.open("r") as f:
        for line in f:
            if line.startswith(f"func {prefix}"):
                test_name = line.split("(")[0].strip().removeprefix("func ")
                test_names.append(test_name)
    return sorted(test_names)


class MultipleResourceNames(ValueError):
    def __init__(self, names: list[str]):
        super().__init__(f"multiple resource names found: {names}")
        self.names = names


def find_tf_resource_name_in_test(path: Path, provider_prefix: str = "mongodbatlas_") -> str:
    candidates: set[str] = {
        match.group(1) for match in re.finditer(rf"=\s\"{provider_prefix}([a-zA-Z0-9_]+)\.?", path.read_text())
    }
    if len(candidates) > 1:
        pkg_name = path.parent.name
        for candidate in candidates:
            if candidate.replace("_", "") == pkg_name:
                return candidate
        logger.warning(f"multiple resource names found in {path}: {candidates}")
        raise MultipleResourceNames(sorted(candidates))
    return candidates.pop() if candidates else ""


def find_pkg_test_names(pkg_path: Path, prefix: str = "Test") -> list[str]:
    test_names = []
    for test_file in pkg_path.glob("*_test.go"):
        test_names.extend(find_test_names(test_file, prefix))
    return sorted(test_names)


def terraform_resource_test_names(
    repo_path: Path, prefix: str = "Test", package_path: str = "internal/service"
) -> dict[str, list[str]]:
    """find all test names in the given package path"""
    pkg_path = terraform_package_path(repo_path, package_path)
    resource_dirs, _ = find_resource_dirs(pkg_path)
    test_names = defaultdict(list)
    for name, pkg_dir in resource_dirs.items():
        for test_file in pkg_dir.glob("*_test.go"):
            test_names[name].extend(find_test_names(test_file, prefix))
    return test_names


def terraform_resources(repo_path: Path, package_path: str = "internal/service") -> list[TFResoure]:
    pkg_path = terraform_package_path(repo_path, package_path)
    resource_dirs, _ = find_resource_dirs(pkg_path)
    resources = []
    for name, pkg_dir in resource_dirs.items():
        test_names = find_pkg_test_names(pkg_dir)
        resources.append(TFResoure(name=name, package_rel_path=str(pkg_dir.relative_to(repo_path)), tests=test_names))
    return resources


def terraform_package_path(repo_path: Path, package_path: str = "internal/service"):
    pkg_path = repo_path / package_path
    if not pkg_path.exists():
        raise ValueError(f"package path not found: {pkg_path}")
    return pkg_path


def find_resource_dirs(pkg_path: Path) -> tuple[dict[str, Path], list[Path]]:
    resource_dirs: dict[str, Path] = {}
    non_resource_dirs: list[Path] = []
    for pkg_dir in pkg_path.iterdir():
        if not pkg_dir.is_dir():
            continue
        if pkg_dir.name == "testdata":
            continue
        found = False
        for test_file in pkg_dir.glob("*_test.go"):
            try:
                if name := find_tf_resource_name_in_test(test_file):
                    resource_dirs[name] = pkg_dir
                    found = True
            except MultipleResourceNames as e:
                for name in e.names:
                    resource_dirs[name] = pkg_dir
                found = True
        if not found:
            non_resource_dirs.append(pkg_dir)
    return resource_dirs, non_resource_dirs


@dataclass
class TFResoure:
    name: str
    package_rel_path: str
    tests: list[str]



from collections import defaultdict
from concurrent.futures import Future, as_completed
from functools import lru_cache
import json
import logging
import os
from pathlib import Path

from ask_shell import new_task, print_to_live, run_pool
from model_lib import dump, parse_model
from pydantic import BaseModel, Field, model_validator
import requests
from rich.markdown import Markdown
import typer
from requests.auth import HTTPDigestAuth
from zero_3rdparty.file_utils import ensure_parents_write_text
from zero_3rdparty.str_utils import ensure_prefix, ensure_suffix, instance_repr

from atlas_init.cli_tf.mock_tf_log import resolve_admin_api_path
from atlas_init.cli_tf.openapi import OpenapiSchema
from atlas_init.settings.env_vars import init_settings
from atlas_init.settings.env_vars_generated import AtlasSettingsWithProject
from atlas_init.settings.env_vars_modules import (
    TFModuleCluster,
    TFModuleFederated_Vars,
    TFModuleProject_Extra,
    TFModuleStream_Instance,
)
from atlas_init.settings.path import load_dotenv
from atlas_init.tf_ext.settings import TfDepSettings

logger = logging.getLogger(__name__)

ALLOWED_MISSING_VARS: set[str] = {
    "alertConfigId",
    "alertId",
    "clientId",
    "cloudProvider",
    "invoiceId",
    # "name",
    "pipelineName",
    "processId",
    "username",
}
ALLOWED_ERROR_CODES: set[str] = {
    "CANNOT_USE_CLUSTER_IN_SERVERLESS_INSTANCE_API",
    "VALIDATION_ERROR",
    "UNEXPECTED_ERROR",
    "CANNOT_USE_NON_FLEX_CLUSTER_IN_FLEX_API",
    "CHECKPOINTS_ONLY_ON_CONTINOUS_BACKUP",
    "INCORRECT_BACKUP_API_ENDPOINT",
}


# export ATLAS_INIT_TEST_SUITES=clusterm10,s3,federated,project,stream_connection
def resolve_path_variables() -> dict[str, str]:
    settings = init_settings()
    env_vars_full = load_dotenv(settings.env_vars_vs_code)
    atlas_settings = AtlasSettingsWithProject(**env_vars_full)
    cluster_settings = TFModuleCluster(**env_vars_full)
    project_settings = TFModuleProject_Extra(**env_vars_full)
    stream_settings = TFModuleStream_Instance(**env_vars_full)
    federated_settings = TFModuleFederated_Vars(**env_vars_full)
    return {
        "orgId": atlas_settings.MONGODB_ATLAS_ORG_ID,
        "cloudProvider": "AWS",
        "federationSettingsId": federated_settings.MONGODB_ATLAS_FEDERATION_SETTINGS_ID,
        "clusterName": cluster_settings.MONGODB_ATLAS_CLUSTER_NAME,
        "name": cluster_settings.MONGODB_ATLAS_CLUSTER_NAME,
        "groupId": atlas_settings.MONGODB_ATLAS_PROJECT_ID,
        "teamId": project_settings.MONGODB_ATLAS_TEAM_ID,
        "tenantName": stream_settings.MONGODB_ATLAS_STREAM_INSTANCE_NAME,
        "apiUserId": atlas_settings.MONGODB_ATLAS_PROJECT_OWNER_ID,
        "username": atlas_settings.MONGODB_ATLAS_USER_EMAIL,
    }


class ApiCall(BaseModel):
    operation_id: str
    path: str
    accept_header: str = "application/vnd.atlas.2023-01-01+json"
    query_args: dict[str, str] = Field(default_factory=dict)

    def __str__(self):
        return instance_repr(self, ["operation_id", "path"])

    def path_with_variables(self, path_variables: dict[str, str]):
        return self.path.format(**path_variables)

    @model_validator(mode="after")
    def check_path_variables(self):
        self.accept_header = ensure_prefix(self.accept_header, "application/vnd.atlas.")
        self.accept_header = ensure_suffix(self.accept_header, "+json")
        return self


class UnresolvedPathsError(Exception):
    def __init__(self, missing_var_paths: dict[str, list[str]]) -> None:
        self.missing_var_paths = missing_var_paths
        missing_vars_formatted = "\n".join(f"{var}: {paths}" for var, paths in missing_var_paths.items())
        super().__init__(f"Failed to resolve path variables:\nMissing vars: {missing_vars_formatted}")


class ApiCalls(BaseModel):
    calls: list[ApiCall] = Field(default_factory=list)
    ignored_calls: list[ApiCall] = Field(default_factory=list)
    path_variables: dict[str, str] = Field(default_factory=resolve_path_variables)
    skip_validation: bool = False

    @model_validator(mode="after")
    def check_path_variables(self):
        if self.skip_validation:
            return self
        missing_vars_paths: dict[str, list[str]] = defaultdict(list)
        ok_calls = []
        for call in self.calls:
            try:
                call.path_with_variables(self.path_variables)
                ok_calls.append(call)
            except KeyError as e:
                missing_vars_paths[str(e).strip("'")].append(f"{call.operation_id} {call.path}")
                self.ignored_calls.append(call)
                continue
        for allowed_missing in sorted(ALLOWED_MISSING_VARS):
            if allowed_missing in missing_vars_paths:
                logger.info(f"Allowed missing variable {allowed_missing}: {missing_vars_paths[allowed_missing]}")
                del missing_vars_paths[allowed_missing]
        if missing_vars_paths:
            raise UnresolvedPathsError(missing_var_paths=missing_vars_paths)
        self.calls = ok_calls
        return self

    def dump_to_dict(self) -> dict:
        return {
            "calls": [call.model_dump(exclude_defaults=True, exclude_unset=True) for call in self.calls],
        }


@lru_cache
def _public_private_key() -> tuple[str, str]:
    public_key = os.environ.get("MONGODB_ATLAS_PUBLIC_KEY")
    private_key = os.environ.get("MONGODB_ATLAS_PRIVATE_KEY")
    if not public_key or not private_key:
        raise ValueError("MONGODB_ATLAS_PUBLIC_KEY and MONGODB_ATLAS_PRIVATE_KEY must be set in environment variables.")
    return public_key, private_key


class APICallError(Exception):
    def __init__(self, api_call: ApiCall, json_response: dict, error: requests.exceptions.HTTPError):
        self.api_call = api_call
        self.json_response = json_response
        super().__init__(f"Failed to make API call {api_call}:\njson={json_response}\n{error}")

    @property
    def error_code(self) -> str:
        return self.json_response.get("errorCode", "")


def call_api(api_call: ApiCall, path_variables: dict[str, str]) -> dict:
    resolved_path = api_call.path_with_variables(path_variables)
    response = requests.get(
        f"https://cloud-dev.mongodb.com/{resolved_path.lstrip('/')}",
        params=api_call.query_args,
        headers={"Accept": api_call.accept_header, "Content-Type": "application/json"},
        auth=HTTPDigestAuth(*_public_private_key()),
        timeout=30,
    )
    try:
        response_json = response.json()
    except requests.exceptions.JSONDecodeError as e:
        logger.error(f"Failed to parse_json {api_call}: {e}")
        response_json = {}
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise APICallError(api_call, response_json, e) from e
    return response_json


class NoSelfLinkError(Exception):
    def __init__(self, json_response: dict) -> None:
        self.json_response = json_response
        super().__init__("No self link found in response")


def parse_href_response(json_response: dict) -> str:
    for ref in json_response.get("links", []):
        if ref.get("rel") == "self":
            return ref.get("href")
    raise NoSelfLinkError(json_response)


def api_config(
    config_path_str: str = typer.Option("", "-p", "--path", help="Path to the API config file"),
    query_args_str: str = typer.Option(
        '{"pageNum": "0", "itemsPerPage": "0"}', "-q", "--query-args", help="Query arguments for the API call"
    ),
    verbose: bool = typer.Option(False, "-v", "--verbose", help="Verbose output"),
):
    query_args: dict[str, str] = json.loads(query_args_str)
    if config_path_str == "":
        with new_task("Find API Calls that use pagination"):
            config_path = dump_config_path(query_args)
    else:
        config_path = Path(config_path_str)
    assert config_path.exists(), f"Config file {config_path} does not exist."
    model = parse_model(config_path, t=ApiCalls)
    total_calls = len(model.calls)
    assert _public_private_key(), "Public and private keys must be set in environment variables."
    path_variables = model.path_variables
    op_id_path_self_qstring: dict[tuple[str, str], str] = {}
    with run_pool(
        task_name="make API calls", max_concurrent_submits=10, threads_used_per_submit=1, total=total_calls
    ) as pool:
        futures: dict[Future, ApiCall] = {
            pool.submit(call_api, api_call, path_variables): api_call for api_call in model.calls
        }
    for future in as_completed(futures):
        api_call = futures[future]
        try:
            result = future.result()
        except APICallError as e:
            if e.error_code in ALLOWED_ERROR_CODES:
                logger.info(f"Allowed error code {e.error_code} in response for {api_call}")
                model.ignored_calls.append(api_call)
                continue
            raise
        except Exception as e:
            logger.error(e)
            continue
        try:
            href = parse_href_response(result)
            op_id_path_self_qstring[(api_call.operation_id, api_call.path)] = href.split("?")[-1]
        except NoSelfLinkError as e:
            logger.error(f"{api_call} did not have a self link in the response:\n{e.json_response}")
            continue
        logger.info(f"API call {api_call} completed successfully with self ref:\n{href}")
        if verbose:
            logger.info(f"Response for {api_call.query_args} was:\n{dump(result, 'pretty_json')}")
    query_args_str = "&".join(f"{key}={value}" for key, value in query_args.items())
    md_report: list[str] = [
        f"# Pagination Report for query_args='{query_args_str}'",
        "",
        "## Checked endpoints",
        "",
        "Operation ID | Path | SelfQueryString",
        "--- | --- | ---",
        *[
            f"{operation_id} | {path} | {self_query_string}"
            for (operation_id, path), self_query_string in op_id_path_self_qstring.items()
        ],
        "",
        "## Ignored endpoints (not checked)",
        "",
        "Operation ID | Path",
        "--- | ---",
        *[f"{call.operation_id} | {call.path}" for call in model.ignored_calls],
    ]
    md_content = "\n".join(md_report)
    md = Markdown(md_content)
    print_to_live(md)
    output_path = TfDepSettings.from_env().pagination_output_path(query_args_str)
    ensure_parents_write_text(output_path, md_content)
    logger.info(f"Pagination report saved to {output_path}")
    return md


def api(
    path: str = typer.Option("-p", "--path", help="Path to the API endpoint"),
    query_string: str = typer.Option("", "-q", "--query-string", help="Query string for the API call"),
):
    assert path, "Path must be provided."
    accept_header = "application/vnd.atlas.2023-01-01+json"
    url = f"https://cloud-dev.mongodb.com/{path.lstrip('/')}?{query_string}"
    logger.info(f"Calling {url}")
    try:
        r = requests.get(
            url,
            headers={"Accept": accept_header, "Content-Type": "application/json"},
            auth=HTTPDigestAuth(*_public_private_key()),
            timeout=30,
        )
        print(r.text)
        r.raise_for_status()
    except requests.exceptions.HTTPError as e:
        print(e)
        print(e.response)


def dump_config_path(query_args: dict[str, str]) -> Path:
    settings = TfDepSettings.from_env()
    latest_api_spec = resolve_admin_api_path()
    model = parse_model(latest_api_spec, t=OpenapiSchema)
    paginated_paths: list[ApiCall] = []
    path_versions = list(model.path_method_api_versions())

    for (path, method, code), versions in path_versions:
        if method != "get" or code != "200":
            continue
        assert len(versions) == 1, f"{path} {method} {code} has multiple versions: {versions}"
        get_method = model.get_method(path)
        if not get_method:
            continue
        parameters = get_method.get("parameters", [])
        for param in parameters:
            if param_ref := param.get("$ref"):
                if param_ref.endswith("itemsPerPage"):
                    version = versions[0].strftime("%Y-%m-%d")
                    paginated_paths.append(
                        ApiCall(
                            path=path,
                            query_args=query_args,
                            accept_header=f"application/vnd.atlas.{version}+json",
                            operation_id=get_method["operationId"],
                        )
                    )
    config_path = settings.api_calls_path
    calls = ApiCalls(
        calls=paginated_paths,
        skip_validation=True,
    )
    calls_yaml = dump(calls.dump_to_dict(), "yaml")
    logger.info(f"Dumped {len(paginated_paths)} API calls to {config_path}")
    ensure_parents_write_text(config_path, calls_yaml)
    return config_path



from atlas_init.settings.env_vars_generated import _EnvVarsGenerated


class TFModuleAws_S3(_EnvVarsGenerated):
    AWS_S3_BUCKET: str


class TFModuleAws_Vars(_EnvVarsGenerated):
    AWS_ACCESS_KEY_ID: str
    AWS_CUSTOMER_MASTER_KEY_ID: str
    AWS_REGION: str
    AWS_REGION_LOWERCASE: str
    AWS_REGION_UPPERCASE: str
    AWS_SECRET_ACCESS_KEY: str


class TFModuleAws_Vpc(_EnvVarsGenerated):
    AWS_SECURITY_GROUP_ID: str
    AWS_SUBNET_ID: str
    AWS_VPC_CIDR_BLOCK: str
    AWS_VPC_ID: str


class TFModuleCfn(_EnvVarsGenerated):
    CFN_EXAMPLE_EXECUTION_ROLE: str
    MONGODB_ATLAS_PRIVATE_API_KEY: str
    MONGODB_ATLAS_PROFILE: str
    MONGODB_ATLAS_PUBLIC_API_KEY: str
    MONGODB_ATLAS_SECRET_PROFILE: str


class TFModuleCloud_Provider(_EnvVarsGenerated):
    AWS_IAM_ROLE_ARN: str
    IAM_ROLE_ID: str


class TFModuleCluster(_EnvVarsGenerated):
    MONGODB_ATLAS_CLUSTER_NAME: str
    MONGODB_ATLAS_CONTAINER_ID: str
    MONGODB_URL: str


class TFModuleFederated_Vars(_EnvVarsGenerated):
    MONGODB_ATLAS_FEDERATED_GROUP_ID: str
    MONGODB_ATLAS_FEDERATED_IDP_ID: str
    MONGODB_ATLAS_FEDERATED_ORG_ID: str
    MONGODB_ATLAS_FEDERATED_SETTINGS_ASSOCIATED_DOMAIN: str
    MONGODB_ATLAS_FEDERATION_SETTINGS_ID: str


class TFModuleProject_Extra(_EnvVarsGenerated):
    MONGODB_ATLAS_ORG_API_KEY_ID: str
    MONGODB_ATLAS_TEAMS_IDS: str
    MONGODB_ATLAS_TEAM_ID: str


class TFModuleStream_Instance(_EnvVarsGenerated):
    MONGODB_ATLAS_STREAM_INSTANCE_ID: str
    MONGODB_ATLAS_STREAM_INSTANCE_NAME: str


class TFModuleVpc_Peering(_EnvVarsGenerated):
    AWS_ACCOUNT_ID: str
    AWS_REGION: str
    AWS_VPC_CIDR_BLOCK: str
    AWS_VPC_ID: str


class TFModuleVpc_Privatelink(_EnvVarsGenerated):
    MONGODB_ATLAS_PRIVATE_ENDPOINT_DNS_NAME: str
    MONGODB_ATLAS_PRIVATE_ENDPOINT_ID: str



from __future__ import annotations

import logging
import os
import re
from collections.abc import Iterable
from fnmatch import fnmatch
from pathlib import Path
from queue import Queue
from tempfile import TemporaryDirectory
from typing import Any, Literal, TypeAlias

from model_lib import Entity, copy_and_validate, parse_model
from pydantic import ConfigDict, Field, model_validator
from zero_3rdparty.enum_utils import StrEnum
from zero_3rdparty.iter_utils import flat_map

from atlas_init.cli_helper.run import run_binary_command_is_ok
from atlas_init.humps import decamelize, pascalize

logger = logging.getLogger(__name__)
INDENT = "  "


class PlanModifier(StrEnum):
    use_state_for_unknown = "UseStateForUnknown"
    diff_suppress_json = "schemafunc.DiffSuppressJSON"


class SchemaAttributeValidator(StrEnum):
    string_is_json = "validate.StringIsJSON"


class SchemaAttribute(Entity):
    model_config = ConfigDict(
        frozen=False,
        validate_assignment=True,
        populate_by_name=True,
        extra="ignore",
    )  # type: ignore

    type: str = ""
    name: str = ""  # populated by the key of the attributes dict
    description: str = ""
    schema_ref: str = ""
    alias: str = ""  # comma separated list of aliases
    is_required: bool = False
    is_optional: bool = False
    is_computed: bool = False
    plan_modifiers: list[PlanModifier] = Field(default_factory=list)
    validators: list[SchemaAttributeValidator] = Field(default_factory=list)
    # not used during dumping but backtrace which parameters are used in the api spec
    parameter_ref: str = ""
    additional_properties: dict[str, Any] = Field(default_factory=dict)

    @property
    def additional_properties_ref(self) -> str:
        if props := self.additional_properties:
            return props.get("$ref", "")
        return ""

    @property
    def tf_name(self) -> str:
        return decamelize(self.name)  # type: ignore

    @property
    def schema_ref_name(self) -> str:
        return self.schema_ref.split("/")[-1] if "/" in self.schema_ref else ""

    @property
    def tf_struct_name(self) -> str:
        return pascalize(self.schema_ref_name) or pascalize(self.name)

    @property
    def aliases(self) -> list[str]:
        return self.alias.split(",") if self.alias else []

    @property
    def is_nested(self) -> bool:
        return self.schema_ref != ""

    def merge(self, other: SchemaAttribute) -> SchemaAttribute:
        # avoid schema_ref when type is different, e.g., setting `pipeline` to a string
        schema_ref = (
            self.schema_ref or other.schema_ref if not self.type or self.type == other.type else self.schema_ref
        )
        return SchemaAttribute(
            type=self.type or other.type,
            name=self.name or other.name,
            description=self.description or other.description,
            schema_ref=schema_ref,
            alias=self.alias or other.alias,
            is_required=self.is_required or other.is_required,
            is_optional=self.is_optional or other.is_optional,
            is_computed=self.is_computed or other.is_computed,
            plan_modifiers=self.plan_modifiers + other.plan_modifiers,
            validators=self.validators + other.validators,
            parameter_ref=self.parameter_ref or other.parameter_ref,
            additional_properties=self.additional_properties | other.additional_properties,
        )

    def set_attribute_type(
        self,
        attr_type: Literal["required", "optional", "computed", "computed_optional"],
    ) -> None:
        match attr_type:
            case "required":
                self.is_required = True
                self.is_computed = False
                self.is_optional = False
            case "optional":
                self.is_optional = True
                self.is_computed = False
                self.is_required = False
            case "computed":
                self.is_computed = True
                self.is_required = False
                self.is_optional = False
            case "computed_optional":
                self.is_optional = True
                self.is_computed = True
                self.is_required = False
            case _:
                raise ValueError(f"Unknown attribute type {attr_type}")


class SkipAttribute(Exception):  # noqa: N818
    def __init__(self, attr_name: str):
        self.attr_name = attr_name
        super().__init__(f"Skipping attribute {attr_name}")


class NewAttribute(Exception):  # noqa: N818
    def __init__(self, attr_name: str):
        self.attr_name = attr_name
        super().__init__(f"New attribute {attr_name}")


def attr_path_matches(attr_path: str, path: str) -> bool:
    return fnmatch(attr_path, path)


class AttributeTypeModifiers(Entity):
    required: list[str] = Field(default_factory=list)
    optional: list[str] = Field(default_factory=list)
    computed: list[str] = Field(default_factory=list)
    computed_optional: set[str] = Field(default_factory=set)

    def set_attribute_type(self, attr: SchemaAttribute, parent_path: str = "") -> SchemaAttribute:
        attr_path = f"{parent_path}.{attr.tf_name}" if parent_path else attr.tf_name
        for attr_type, names in self.model_dump().items():
            for path_matcher in names:
                if attr_path_matches(attr_path, path_matcher):
                    logger.info(f"Setting attribute {attr_path} to {attr_type}")
                    attr.set_attribute_type(attr_type)  # type: ignore
                    return attr
        if attr.name in self.required:
            attr.is_required = True
            attr.is_computed = False
            attr.is_optional = False
        if attr.is_computed and attr.is_required:
            logger.warning(f"Attribute {attr.name} cannot be both computed and required, using required")
            attr.is_computed = False
        if attr.is_optional and attr.is_required:
            raise ValueError(f"Attribute {attr.name} cannot be both optional and required")
        if not attr.is_computed and not attr.is_required and not attr.is_optional:
            logger.warning(
                f"Attribute {attr.name} is neither read_only, required, nor optional, using computed-optional"
            )
            attr.is_optional = True
            attr.is_computed = True
        return attr


class SDKModelExample(Entity):
    name: str
    examples: list[str] = Field(default_factory=list)


class SDKConversion(Entity):
    sdk_start_refs: list[SDKModelExample] = Field(default_factory=list)

    def __bool__(self) -> bool:
        return bool(self.sdk_start_refs)


class Discriminator(Entity):
    mapping: dict[str, str] = Field(default_factory=dict)
    property_name: str = Field(alias="propertyName")


class OneOf(Entity):
    ref: str = Field(alias="$ref", default="")


class AllOf(Entity):
    ref: str = Field(alias="$ref", default="")
    properties: dict[str, Any] = Field(default_factory=dict)

    @property
    def nested_refs(self) -> set[str]:
        refs = set()
        for prop, prop_value in self.properties.items():
            if isinstance(prop_value, dict):
                if ref := prop_value.get("$ref"):
                    refs.add(ref)
        return refs


class SchemaResource(Entity):
    name: str = ""  # populated by the key of the resources dict
    description: str = ""
    attributes: dict[str, SchemaAttribute] = Field(default_factory=dict)
    attributes_skip: set[str] = Field(default_factory=set)
    paths: list[str] = Field(default_factory=list)
    attribute_type_modifiers: AttributeTypeModifiers = Field(default_factory=AttributeTypeModifiers)
    conversion: SDKConversion = Field(default_factory=SDKConversion)
    discriminator: Discriminator | None = None
    one_of: list[OneOf] = Field(default_factory=list)
    all_of: list[AllOf] = Field(default_factory=list)

    def extra_refs(self) -> set[str]:
        return (
            {one_of.ref for one_of in self.one_of if one_of.ref}
            | {all_of.ref for all_of in self.all_of if all_of.ref}
            | {ref for ref in flat_map(all_of.nested_refs for all_of in self.all_of) if ref}
        )

    @model_validator(mode="after")
    def set_attribute_names(self):
        for name, attr in self.attributes.items():
            attr.name = name
        return self

    @property
    def nested_refs(self) -> set[str]:
        return {attr.schema_ref for attr in self.attributes.values() if attr.is_nested}

    def lookup_tf_name(self, tf_name: str) -> SchemaAttribute:
        for name, attr in self.attributes.items():
            if tf_name == decamelize(name):
                return attr
            for alias in attr.aliases:
                if tf_name == decamelize(alias):
                    return attr
        raise ValueError(f"Attribute {tf_name} not found in resource {self.name}")

    def lookup_attribute(self, name: str) -> SchemaAttribute:
        if name in self.attributes_skip:
            raise SkipAttribute(name)
        if found := self.attributes.get(name):
            return found
        for attr in self.attributes.values():
            if name in attr.aliases:
                return attr
        raise NewAttribute(name)

    def sorted_attributes(self) -> list[SchemaAttribute]:
        return sorted(self.attributes.values(), key=lambda a: a.name)


class OpenAPIChanges(Entity):
    schema_prefix_removal: list[str] = Field(default_factory=list)


class SchemaV2(Entity):
    attributes_skip: set[str] = Field(default_factory=set)
    openapi_changes: OpenAPIChanges = Field(default_factory=OpenAPIChanges)
    resources: dict[str, SchemaResource] = Field(default_factory=dict)
    ref_resources: dict[str, SchemaResource] = Field(default_factory=dict)

    def ref_resource(self, ref: str, use_name: str = "") -> SchemaResource:
        if ref not in self.ref_resources:
            raise ValueError(f"Resource {ref} not found in ref_resources")
        resource = self.ref_resources[ref]
        return copy_and_validate(resource, name=use_name) if use_name else resource

    @model_validator(mode="after")
    def set_resource_names(self):
        for name, resource in self.resources.items():
            resource.name = name
        return self

    @model_validator(mode="after")
    def add_ignored_attributes(self):
        for resource in self.resources.values():
            resource.attributes_skip |= self.attributes_skip
        return self

    def reset_attributes_skip(self) -> None:
        self.attributes_skip.clear()
        for resource in self.resources.values():
            resource.attributes_skip.clear()


def parse_schema(path: Path) -> SchemaV2:
    return parse_model(path, t=SchemaV2)


def generate_resource_go_resource_schema(schema: SchemaV2, resource_name: str) -> str:
    if resource_name not in schema.resources:
        raise ValueError(f"Resource {resource_name} not found in schema")
    resource = schema.resources[resource_name]
    return generate_go_resource_schema(schema, resource)


def generate_resource_go_schemas(schema: SchemaV2) -> Iterable[str]:
    for name, resource in schema.resources.items():
        logger.info(f"Generating Go schema for {name}")
        yield generate_go_resource_schema(schema, resource)


def package_name(resource_name: str) -> str:
    return resource_name.replace("_", "").lower()


def indent(level: int, line: str) -> str:
    return INDENT * level + line


admin_version = os.getenv("ATLAS_SDK_VERSION", "v20241023001")

_import_urls = [
    "context",
    "github.com/hashicorp/terraform-plugin-framework/attr",
    "github.com/hashicorp/terraform-plugin-framework/diag",
    "github.com/hashicorp/terraform-plugin-framework/resource/schema",
    "github.com/hashicorp/terraform-plugin-framework/resource/schema/planmodifier",
    "github.com/hashicorp/terraform-plugin-framework/resource/schema/stringplanmodifier",
    "github.com/hashicorp/terraform-plugin-framework/schema/validator",
    "github.com/hashicorp/terraform-plugin-framework/types",
    "github.com/hashicorp/terraform-plugin-framework/types/basetypes",
    "github.com/hashicorp/terraform-plugin-framework-timeouts/resource/timeouts",
    "github.com/mongodb/terraform-provider-mongodbatlas/internal/common/conversion",
    "github.com/mongodb/terraform-provider-mongodbatlas/internal/common/schemafunc",
    "github.com/mongodb/terraform-provider-mongodbatlas/internal/common/validate",
    f"go.mongodb.org/atlas-sdk/{admin_version}/admin",
]
_import_urls_dict = {url.split("/")[-1]: url for url in _import_urls}

package_usage_pattern = re.compile(r"(?P<package_name>[\w\d_]+)\.(?P<package_func>[\w\d_]+)")

_variable_names = set()


def add_go_variable_names(names: Iterable[str]) -> None:
    _variable_names.update(names)


_variable_suffixes = (
    "Model",
    "ObjectType",
    "ObjType",
)


def extend_import_urls(import_urls: set[str], code_lines: list[str]) -> None:
    for line in code_lines:
        for match in package_usage_pattern.finditer(line):
            package_name = match.group("package_name")
            if package_name in _variable_names:
                continue
            if package_name.endswith(_variable_suffixes):
                continue
            if package_name in _import_urls_dict:
                import_urls.add(_import_urls_dict[package_name])
            else:
                err_msg = f"Unknown package '{package_name}' used in line {line}"
                raise ValueError(err_msg)


def import_lines(import_urls: set[str]) -> list[str]:
    stdlib_imports = {url for url in import_urls if "." not in url}
    pkg_imports = import_urls - stdlib_imports
    imports = sorted(stdlib_imports)
    if imports:
        imports.append("")
    imports.extend(sorted(pkg_imports))
    return [
        "import (",
        *[indent(1, f'"{url}"') if url else "" for url in imports],
        ")",
        "",
    ]


def generate_go_resource_schema(schema: SchemaV2, resource: SchemaResource) -> str:
    func_lines = resource_schema_func(schema, resource)
    object_type_lines = resource_object_type_lines(schema, resource)
    import_urls = set()
    extend_import_urls(import_urls, func_lines)
    extend_import_urls(import_urls, object_type_lines)
    unformatted = "\n".join(
        [
            f"package {package_name(resource.name)}",
            "",
            *import_lines(import_urls),
            "",
            *func_lines,
            "",
            *object_type_lines,
        ]
    )
    return go_fmt(resource.name, unformatted)


def go_fmt(name: str, unformatted: str) -> str:
    with TemporaryDirectory() as temp_dir:
        filename = f"{name}.go"
        result_file = Path(temp_dir) / filename
        result_file.write_text(unformatted)
        if not run_binary_command_is_ok("go", f"fmt {filename}", cwd=Path(temp_dir), logger=logger):
            logger.warning(f"go file unformatted:\n{unformatted}")
            raise ValueError(f"Failed to format {result_file}")
        return result_file.read_text()


def resource_schema_func(schema: SchemaV2, resource: SchemaResource) -> list[str]:
    func_lines = [
        "func ResourceSchema(ctx context.Context) schema.Schema {",
        indent(1, "return schema.Schema{"),
        indent(2, "Attributes: map[string]schema.Attribute{"),
    ]
    for attr in resource.sorted_attributes():
        func_lines.extend(generate_go_attribute_schema_lines(schema, attr, 3, [resource]))
    func_lines.extend((indent(2, "},"), indent(1, "}"), "}"))
    return func_lines


_attr_schema_types = {
    "string": "schema.StringAttribute",
}
_attr_nested_schema_types = {
    "object": "schema.SingleNestedAttribute",
    "array": "schema.ListNestedAttribute",
}


def attribute_header(attr: SchemaAttribute) -> str:
    if attr.is_nested:
        err_msg = f"Unknown nested attribute type: {attr.type}"
        schema_types = _attr_nested_schema_types
    else:
        err_msg = f"Unknown attribute type: {attr.type}"
        schema_types = _attr_schema_types
    header = schema_types.get(attr.type)
    if header is None:
        raise NotImplementedError(err_msg)
    return header


def plan_modifier_call(modifier: PlanModifier, default_pkg_name: str) -> str:
    if "." in modifier:
        return f"{modifier}(),"
    return f"{default_pkg_name}.{pascalize(modifier)}(),"


def plan_modifiers_lines(attr: SchemaAttribute, line_indent: int) -> list[str]:
    # sourcery skip: reintroduce-else, swap-if-else-branches, use-named-expression
    plan_modifiers = attr.plan_modifiers
    if not plan_modifiers:
        return []
    modifier_package = {"string": "stringplanmodifier"}.get(attr.type)
    if not modifier_package:
        raise NotImplementedError
    modifier_header = {
        "string": "planmodifier.String",
    }.get(attr.type)
    if not modifier_header:
        raise NotImplementedError
    return [
        indent(line_indent, f"PlanModifiers: []{modifier_header}{{"),
        *[indent(line_indent + 1, plan_modifier_call(modifier, modifier_package)) for modifier in plan_modifiers],
        indent(line_indent, "},"),
    ]


def validate_call(validator: SchemaAttributeValidator) -> str:
    if "." not in validator:
        raise NotImplementedError
    return f"{validator}(),"


def validate_attribute_lines(attr: SchemaAttribute, line_indent: int) -> list[str]:
    # sourcery skip: reintroduce-else, swap-if-else-branches, use-named-expression
    if not attr.validators or attr.is_computed:
        return []
    validator_header = {
        "string": "validator.String",
    }.get(attr.type)
    if not validator_header:
        raise NotImplementedError
    return [
        indent(line_indent, f"Validators: []{validator_header}{{"),
        *[indent(line_indent + 1, validate_call(validator)) for validator in attr.validators],
        indent(line_indent, "},"),
    ]


def generate_go_attribute_schema_lines(
    schema: SchemaV2,
    attr: SchemaAttribute,
    line_indent: int,
    parent_resources: list[SchemaResource],
) -> list[str]:
    parent_path = ".".join(parent.name for parent in parent_resources[1:])
    parent_resources[0].attribute_type_modifiers.set_attribute_type(attr, parent_path)
    attr_name = attr.tf_name
    lines = [indent(line_indent, f'"{attr_name}": {attribute_header(attr)}{{')]
    if desc := attr.description or attr.is_nested and (desc := schema.ref_resource(attr.schema_ref).description):
        lines.append(indent(line_indent + 1, f'Description: "{desc.replace("\n", "\\n")}",'))
        lines.append(indent(line_indent + 1, f'MarkdownDescription: "{desc.replace("\n", "\\n")}",'))
    if attr.is_required:
        lines.append(indent(line_indent + 1, "Required: true,"))
    if attr.is_optional:
        lines.append(indent(line_indent + 1, "Optional: true,"))
    if attr.is_computed:
        lines.append(indent(line_indent + 1, "Computed: true,"))
    if attr.validators:
        lines.extend(validate_attribute_lines(attr, line_indent + 1))
    if attr.plan_modifiers:
        lines.extend(plan_modifiers_lines(attr, line_indent + 1))
    if attr.is_nested:
        nested_attr = schema.ref_resource(attr.schema_ref, use_name=attr_name)
        if attr.type == "array":
            lines.append(indent(line_indent + 1, "NestedObject: schema.NestedAttributeObject{"))
            lines.extend(generate_nested_attribute_schema_lines(schema, line_indent + 1, parent_resources, nested_attr))
            lines.append(indent(line_indent + 1, "},"))
        else:
            lines.extend(generate_nested_attribute_schema_lines(schema, line_indent + 1, parent_resources, nested_attr))
    lines.append(indent(line_indent, "},"))
    return lines


def generate_nested_attribute_schema_lines(
    schema: SchemaV2, line_indent: int, parent_resources: list[SchemaResource], nested_attr: SchemaResource
) -> list[str]:
    lines = [indent(line_indent, "Attributes: map[string]schema.Attribute{")]
    for nes in nested_attr.attributes.values():
        lines.extend(generate_go_attribute_schema_lines(schema, nes, line_indent + 1, [*parent_resources, nested_attr]))
    lines.append(indent(line_indent, "},"))
    return lines


ResourceTypes: TypeAlias = Literal["rs", "ds", "dsp", ""]


def as_struct_name(resource_name: str, resource_type: ResourceTypes) -> str:
    return f"TF{pascalize(resource_name)}{resource_type.upper()}Model"


def struct_def(tf_name: str, resource_type: ResourceTypes) -> str:
    name = as_struct_name(tf_name, resource_type)
    return f"type {name} struct {{"


_tpf_types = {
    "string": "String",
    "int": "Int",
    "bool": "Bool",
    "map": "Map",
    "list": "List",
    "array": "List",
    "object": "Object",
}


def as_tpf_type(type_name: str) -> str:
    if tpf_type := _tpf_types.get(type_name):
        return tpf_type
    raise ValueError(f"Don't know how to convert {type_name} to TPF type")


def struct_field_line(attr: SchemaAttribute) -> str:
    tpf_type = as_tpf_type(attr.type)
    struct_field_name = pascalize(attr.tf_name).replace("Id", "ID").replace("Db", "DB")  # type: ignore
    return f'{struct_field_name} types.{tpf_type} `tfsdk:"{attr.tf_name}"`'


def as_object_type_name(attr: SchemaAttribute) -> str:
    tpf_type = as_tpf_type(attr.type)
    return f"{tpf_type}Type"


def custom_object_type_name(attr: SchemaAttribute) -> str:
    return f"{pascalize(attr.tf_struct_name)}ObjectType"


def object_type_def(attr: SchemaAttribute) -> str:
    return f"var {custom_object_type_name(attr)} = types.ObjectType{{AttrTypes: map[string]attr.Type{{"


def object_type_field_line(attr: SchemaAttribute) -> str:
    if attr.is_nested:
        return f'"{attr.tf_name}": {custom_object_type_name(attr)},'
    object_type_name = as_object_type_name(attr)
    return f'"{attr.tf_name}": types.{object_type_name},'


_used_refs: set[str] = set()


def resource_object_type_lines(schema: SchemaV2, resource: SchemaResource) -> list[str]:
    nested_attributes: Queue[SchemaAttribute] = Queue()
    lines = [
        struct_def(resource.name, "rs"),
        *[indent(1, struct_field_line(attr)) for attr in resource.sorted_attributes()],
        "}",
        "",
    ]
    for attr in resource.sorted_attributes():
        if attr.is_nested:
            nested_attributes.put(attr)
    while not nested_attributes.empty():
        nested_attr = nested_attributes.get()
        schema_ref = nested_attr.schema_ref
        if schema_ref in _used_refs:
            continue
        _used_refs.add(schema_ref)
        nested_resource = schema.ref_resource(schema_ref, use_name=nested_attr.tf_name)
        logger.info(f"creating struct for nested attribute %s, ref={schema_ref}", nested_attr.name)
        lines.extend(
            [
                struct_def(nested_attr.tf_struct_name, ""),
                *[indent(1, struct_field_line(attr)) for attr in nested_resource.sorted_attributes()],
                "}",
                "",
            ]
        )
        lines.extend(
            [
                object_type_def(nested_attr),
                *[indent(1, object_type_field_line(attr)) for attr in nested_resource.sorted_attributes()],
                "}}",
                "",
            ]
        )
        for attr in nested_resource.sorted_attributes():
            if attr.is_nested:
                nested_attributes.put(attr)
    return lines



import logging
import os
from collections.abc import Callable
from functools import lru_cache
from pathlib import Path

import requests
from github import Auth, Github
from github.Repository import Repository
from github.WorkflowJob import WorkflowJob
from github.WorkflowRun import WorkflowRun
from github.WorkflowStep import WorkflowStep
from zero_3rdparty import datetime_utils, file_utils

from atlas_init.repos.path import (
    GH_OWNER_TERRAFORM_PROVIDER_MONGODBATLAS,
)
from atlas_init.settings.env_vars import init_settings

logger = logging.getLogger(__name__)

GH_TOKEN_ENV_NAME = "GH_TOKEN"  # noqa: S105 #nosec
REQUIRED_GH_ENV_VARS = [GH_TOKEN_ENV_NAME]
MAX_DOWNLOADS = 5


@lru_cache
def get_auth() -> Auth.Auth:
    token = os.environ[GH_TOKEN_ENV_NAME]
    return Auth.Token(token)


@lru_cache
def get_repo(repo_id: str) -> Repository:
    auth = get_auth()
    g = Github(auth=auth)
    logger.info(f"logged in as: {g.get_user().login}")
    return g.get_repo(repo_id)


def tf_repo() -> Repository:
    return get_repo(GH_OWNER_TERRAFORM_PROVIDER_MONGODBATLAS)


def download_job_safely(workflow_dir: Path, job: WorkflowJob) -> Path | None:
    if job.conclusion in {"skipped", "cancelled", None}:
        logger.debug(f"not downloading job: {job.html_url}, conclusion: {job.conclusion}")
        return None
    path = logs_file(workflow_dir, job)
    job_summary = f"found test job: {job.name}, attempt {job.run_attempt}, {job.created_at}, url: {job.html_url}"
    if path.exists():
        logger.debug(f"{job_summary} exist @ {path}")
        return path
    logger.debug(f"{job_summary}\n\t\t downloading to {path}")
    try:
        logs_response = requests.get(job.logs_url(), timeout=60)
        logs_response.raise_for_status()
    except Exception as e:  # noqa: BLE001
        logger.warning(f"failed to download logs for {job.html_url}, e={e!r}")
        return None
    file_utils.ensure_parents_write_text(path, logs_response.text)
    return path


def logs_dir() -> Path:
    return init_settings().github_ci_run_logs


def summary_dir(summary_name: str) -> Path:
    return init_settings().github_ci_summary_dir / summary_name


def workflow_logs_dir(workflow: WorkflowRun) -> Path:
    dt = workflow.created_at
    date_str = datetime_utils.get_date_as_rfc3339_without_time(dt)
    workflow_name = Path(workflow.path).stem
    return logs_dir() / f"{date_str}/{workflow.id}_{workflow_name}"


def logs_file(workflow_dir: Path, job: WorkflowJob) -> Path:
    if job.run_attempt != 1:
        workflow_dir = workflow_dir.with_name(f"{workflow_dir.name}_attempt{job.run_attempt}")
    filename = f"{job.id}_" + job.name.replace(" ", "").replace("/", "_").replace("__", "_") + ".log"
    return workflow_dir / filename


def as_test_group(job_name: str) -> str:
    """tests-1.8.x-latest / tests-1.8.x-latest-dev / config"""
    return "" if "/" not in job_name else job_name.split("/")[-1].strip()


def include_test_jobs(test_group: str = "") -> Callable[[WorkflowJob], bool]:
    def inner(job: WorkflowJob) -> bool:
        job_name = job.name
        if test_group:
            return is_test_job(job_name) and as_test_group(job_name) == test_group
        return is_test_job(job.name)

    return inner


def is_test_job(job_name: str) -> bool:
    """
    >>> is_test_job("tests-1.8.x-latest / tests-1.8.x-latest-dev / config")
    True
    """
    if "-before" in job_name or "-after" in job_name:
        return False
    return "tests-" in job_name and not job_name.endswith(("get-provider-version", "change-detection"))


def select_step_and_log_content(job: WorkflowJob, logs_path: Path) -> tuple[int, list[str]]:
    full_text = logs_path.read_text()
    step = test_step(job.steps)
    last_step_start = current_step_start = 1
    # there is always an extra setup job step, so starting at 1
    current_step = 1
    lines = full_text.splitlines()
    for line_index, line in enumerate(lines, 0):
        if "##[group]Run " in line:
            current_step += 1
            last_step_start, current_step_start = current_step_start, line_index
            if current_step == step + 1:
                return step, lines[last_step_start:current_step_start]
    assert step == current_step, f"didn't find enough step in logs for {job.html_url}"
    return step, lines[current_step_start:]


def test_step(steps: list[WorkflowStep]) -> int:
    for i, step in enumerate(steps, 1):
        name_lower = step.name.lower()
        if "acceptance test" in name_lower and "mocked" not in name_lower:
            return i
    last_step = len(steps)
    logger.warning(f"using {last_step} as final step, unable to find 'test' in {steps}")
    return last_step



import json
import logging

from atlas_init.cli_tf.debug_logs import SDKRoundtrip, parsed
from atlas_init.cli_tf.debug_logs_test_data import RTModifier

logger = logging.getLogger(__name__)


def add_label_tags(rt: SDKRoundtrip):
    logger.info(f"Adding labels and tags to {rt.id}")
    request = rt.request
    req_dict, req_list, req_bool = parsed(request.text)
    response = rt.response
    resp_dict, resp_list, req_bool = parsed(response.text)
    if resp_list or req_list or req_bool is not None:
        return
    resp_dict = resp_dict or {}
    req_dict = req_dict or {}
    for extra_field in ["labels", "tags"]:
        if extra_field not in resp_dict:
            resp_dict[extra_field] = []
        if extra_field not in req_dict:
            req_dict[extra_field] = []
    request.text = json.dumps(req_dict, indent=1, sort_keys=True)
    response.text = json.dumps(resp_dict, indent=1, sort_keys=True)


cluster_modifier = RTModifier(
    version="2024-08-05",
    method="POST",
    path="/api/atlas/v2/groups/{groupId}/clusters",
    modification=add_label_tags,
)


def package_modifiers(pkg_name: str) -> list[RTModifier]:
    # sourcery skip: assign-if-exp, reintroduce-else
    if pkg_name == "advancedcluster":
        return [cluster_modifier]
    return []


def package_skip_suffixes(pkg_name: str) -> list[str]:
    # sourcery skip: assign-if-exp, reintroduce-else
    if pkg_name == "resourcepolicy":
        return [":validate"]
    return []


def package_must_substrings(pkg_name: str) -> list[str]:
    # sourcery skip: assign-if-exp, reintroduce-else
    if pkg_name == "advancedcluster":
        return ["/clusters"]
    return []



import logging
from collections import defaultdict
from functools import total_ordering
from pathlib import Path

import typer
from model_lib import Entity, Event, dump, parse_payload
from pydantic import Field

from atlas_init.cli_helper.run import run_binary_command_is_ok
from atlas_init.cli_tf.hcl.modifier import (
    BLOCK_TYPE_OUTPUT,
    BLOCK_TYPE_VARIABLE,
    update_descriptions,
)

logger = logging.getLogger(__name__)


class UpdateExamples(Entity):
    examples_base_dir: Path
    var_descriptions: dict[str, str]
    output_descriptions: dict[str, str] = Field(default_factory=dict)
    skip_tf_fmt: bool = False


@total_ordering
class TFConfigDescriptionChange(Event):
    path: Path
    name: str
    before: str
    after: str
    block_type: str

    @property
    def changed(self) -> bool:
        return self.after not in ("", self.before)

    def __lt__(self, other) -> bool:
        if not isinstance(other, TFConfigDescriptionChange):
            raise TypeError
        return (self.path, self.name) < (other.path, other.name)


class UpdateExamplesOutput(Entity):
    before_var_descriptions: dict[str, str] = Field(default_factory=dict)
    before_output_descriptions: dict[str, str] = Field(default_factory=dict)
    changes: list[TFConfigDescriptionChange] = Field(default_factory=list)


def update_examples(event_in: UpdateExamples) -> UpdateExamplesOutput:
    changes = []
    existing_var_descriptions = update_block_descriptions(
        event_in.examples_base_dir,
        changes,
        event_in.var_descriptions,
        BLOCK_TYPE_VARIABLE,
    )
    existing_output_descriptions = update_block_descriptions(
        event_in.examples_base_dir,
        changes,
        event_in.output_descriptions,
        BLOCK_TYPE_OUTPUT,
    )
    if event_in.skip_tf_fmt:
        logger.info("skipping terraform fmt")
    else:
        assert run_binary_command_is_ok("terraform", "fmt -recursive", cwd=event_in.examples_base_dir, logger=logger), (
            "terraform fmt failed"
        )
    return UpdateExamplesOutput(
        before_var_descriptions=flatten_descriptions(existing_var_descriptions),
        before_output_descriptions=flatten_descriptions(existing_output_descriptions),
        changes=sorted(changes),
    )


def flatten_descriptions(descriptions: dict[str, list[str]]) -> dict[str, str]:
    return {
        key: "\n".join(desc for desc in sorted(set(descriptions)) if desc != "")
        for key, descriptions in descriptions.items()
    }


def update_block_descriptions(
    base_dir: Path,
    changes: list[TFConfigDescriptionChange],
    new_names: dict[str, str],
    block_type: str,
):
    all_existing_descriptions = defaultdict(list)
    in_files = sorted(base_dir.rglob("*.tf"))
    for tf_file in in_files:
        logger.info(f"looking for {block_type} in {tf_file}")
        new_tf, existing_descriptions = update_descriptions(tf_file, new_names, block_type=block_type)
        if not existing_descriptions:  # probably no variables in the file
            continue
        for name, descriptions in existing_descriptions.items():
            changes.extend(
                TFConfigDescriptionChange(
                    path=tf_file,
                    name=name,
                    before=description,
                    after=new_names.get(name, ""),
                    block_type=block_type,
                )
                for description in descriptions
            )
            all_existing_descriptions[name].extend(descriptions)
        if tf_file.read_text() == new_tf:
            logger.debug(f"no {block_type} changes for {tf_file}")
            continue
        tf_file.write_text(new_tf)
    return all_existing_descriptions


def update_example_cmd(
    examples_base_dir: Path = typer.Argument(
        ..., help="Directory containing *.tf files (can have many subdirectories)"
    ),
    var_descriptions: Path = typer.Option("", "--vars", help="Path to a JSON/yaml file with variable descriptions"),
    output_descriptions: Path = typer.Option("", "--outputs", help="Path to a JSON/yaml file with output descriptions"),
    skip_log_existing: bool = typer.Option(False, help="Log existing descriptions"),
    skip_log_changes: bool = typer.Option(False, help="Log variable updates"),
):
    var_descriptions_dict = parse_payload(var_descriptions) if var_descriptions else {}
    output_descriptions_dict = parse_payload(output_descriptions) if output_descriptions else {}
    event = UpdateExamples(
        examples_base_dir=examples_base_dir,
        var_descriptions=var_descriptions_dict,  # type: ignore
        output_descriptions=output_descriptions_dict,  # type: ignore
    )
    output = update_examples(event)
    if not skip_log_changes:
        for change in output.changes:
            if change.changed:
                logger.info(f"{change.path}({change.block_type}) {change.name}: {change.before} -> {change.after}")
    if not skip_log_existing:
        existing_var_yaml = dump(output.before_var_descriptions, "yaml")
        logger.info(f"Existing Variables:\n{existing_var_yaml}")
        existing_output_yaml = dump(output.before_output_descriptions, "yaml")
        logger.info(f"Existing Outputs:\n{existing_output_yaml}")



import logging

import requests
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed
from zero_3rdparty.id_creator import simple_id

from atlas_init.settings.env_vars import env_vars_cls_or_none, init_settings
from atlas_init.settings.env_vars_generated import (
    AtlasSettingsWithProject,
    AWSSettings,
    RealmSettings,
)
from atlas_init.settings.env_vars_modules import TFModuleCluster
from atlas_init.settings.path import dump_dotenv
from atlas_init.typer_app import app_command

logger = logging.getLogger(__name__)


@app_command()
def trigger_app():
    create_realm_app()


def create_realm_app():
    settings = init_settings(AtlasSettingsWithProject, TFModuleCluster, AWSSettings)
    atlas_settings = AtlasSettingsWithProject.from_env()
    cluster_settings = TFModuleCluster.from_env()
    aws_settings = AWSSettings.from_env()
    project_id = atlas_settings.MONGODB_ATLAS_PROJECT_ID
    base_url = atlas_settings.realm_url
    cluster_name = cluster_settings.MONGODB_ATLAS_CLUSTER_NAME
    auth_headers = login_to_realm(settings, base_url)
    realm_settings = env_vars_cls_or_none(RealmSettings, dotenv_path=settings.env_vars_trigger)
    if realm_settings and function_exists(
        base_url,
        auth_headers,
        project_id,
        realm_settings.MONGODB_REALM_APP_ID,
        realm_settings.MONGODB_REALM_FUNCTION_ID,
    ):
        logger.info(f"function {realm_settings.MONGODB_REALM_FUNCTION_NAME} already exists ✅")
        settings.include_extra_env_vars_in_vscode(realm_settings.model_dump())
        return
    logger.info("creating new realm app")
    if apps := list_apps(base_url, auth_headers, project_id):
        logger.info(f"got apps: {apps}")
        app_id = apps[0]["_id"]
    else:
        logger.info("no apps found, creating one")
        app = create_app(base_url, auth_headers, project_id, cluster_name, aws_settings.AWS_REGION)
        logger.info(f"created app: {app}")
        app_id = app["_id"]
    logger.info(f"using app_id: {app_id}")
    suffix = simple_id(length=5)
    service = create_service(base_url, auth_headers, project_id, cluster_name, app_id, suffix)
    logger.info(f"new service: {service}")
    service_id = service["_id"]
    logger.info(f"using service_id: {service_id}")
    func_response = create_function(base_url, auth_headers, project_id, app_id, suffix)
    logger.info(f"new function: {func_response}")
    func_id = func_response["_id"]
    func_name = func_response["name"]
    logger.info(f"using func_id: {func_id}")
    realm_settings = RealmSettings(
        MONGODB_REALM_APP_ID=app_id,
        MONGODB_REALM_SERVICE_ID=service_id,
        MONGODB_REALM_FUNCTION_ID=func_id,
        MONGODB_REALM_FUNCTION_NAME=func_name,
        MONGODB_REALM_BASE_URL=base_url,
    )
    extra_env_vars = realm_settings.model_dump()
    dump_dotenv(settings.env_vars_trigger, extra_env_vars)
    logger.info(f"done {settings.env_vars_trigger} created with trigger env-vars ✅")
    settings.include_extra_env_vars_in_vscode(extra_env_vars)


def login_to_realm(settings, base_url):
    login_req = {
        "username": settings.MONGODB_ATLAS_PUBLIC_KEY,
        "apiKey": settings.MONGODB_ATLAS_PRIVATE_KEY,
    }
    token_response = _request_post_call(
        f"{base_url}api/admin/v3.0/auth/providers/mongodb-cloud/login",
        data=login_req,
        headers={"Accept": "application/json", "Content-Type": "application/json"},
        timeout=10,
    )
    access_token = token_response["access_token"]
    logger.debug(f"token: {access_token}")
    return {"Authorization": f"Bearer {access_token}"}


def list_apps(base_url: str, auth_headers: dict[str, str], project_id: str) -> list[dict]:
    existing_apps_response = requests.get(
        f"{base_url}api/admin/v3.0/groups/{project_id}/apps",
        headers=auth_headers,
        timeout=10,
    )
    existing_apps_response.raise_for_status()
    apps = existing_apps_response.json()
    assert isinstance(apps, list), f"expected list, got: {apps!r}"
    return apps


# https://www.mongodb.com/docs/atlas/app-services/apps/deployment-models-and-regions/#cloud-deployment-regions
_cloud_deployment_regions = {
    "aws-eu-west-1": "IE",
    "aws-us-west-2": "US-OR",
    "aws-ap-southeast-2": "AU",
    "aws-us-east-1": "US-VA",
}


def create_app(
    base_url: str,
    auth_headers: dict[str, str],
    project_id: str,
    cluster_name: str,
    aws_region: str,
) -> dict:
    provider_region = f"aws-{aws_region}"
    location = _cloud_deployment_regions.get(provider_region)
    if not location:
        raise ValueError(
            f"unknown location for provider_region: {provider_region}, only supports: {_cloud_deployment_regions}"
        )
    create_app_req = {
        "name": "atlas-init-app",
        "location": location,
        "deployment_model": "GLOBAL",
        "environment": "production",
        "provider_region": provider_region,
        "data_source": {
            "name": "mongodb-atlas",
            "type": "mongodb-atlas",
            "config": {
                "clusterName": cluster_name,
                "readPreference": "primary",
                "wireProtocolEnabled": True,
            },
        },
    }
    app_response = _request_post_call(
        f"{base_url}api/admin/v3.0/groups/{project_id}/apps",
        data=create_app_req,
        headers=auth_headers,
        timeout=10,
        log_data_on_failure=True,
    )
    assert isinstance(app_response, dict), f"expected dict, got: {app_response!r}"
    return app_response


def create_service(
    base_url: str,
    auth_headers: dict[str, str],
    project_id: str,
    cluster_name: str,
    app_id: str,
    suffix: str,
) -> dict:
    create_service_req = {
        "name": f"atlas-init-{suffix}",
        "type": "mongodb-atlas",
        "config": {
            "clusterName": cluster_name,
            "readPreference": "primary",
            "wireProtocolEnabled": True,
        },
    }
    service = _request_post_call(
        f"{base_url}api/admin/v3.0/groups/{project_id}/apps/{app_id}/services",
        data=create_service_req,
        headers=auth_headers,
        timeout=10,
        log_data_on_failure=True,
    )
    assert isinstance(service, dict), f"expected dict, got: {service}"
    return service


def create_function(
    base_url: str,
    auth_headers: dict[str, str],
    project_id: str,
    app_id: str,
    suffix: str,
) -> dict:
    create_func_req = {
        "can_evaluate": {},
        "name": f"testfunc-{suffix}",
        "private": True,
        "source": 'exports = function(changeEvent) {console.log("New Document Inserted")};',
        "run_as_system": True,
    }
    func = _request_post_call(
        f"{base_url}api/admin/v3.0/groups/{project_id}/apps/{app_id}/functions",
        data=create_func_req,
        headers=auth_headers,
        timeout=10,
        log_data_on_failure=True,
    )
    assert isinstance(func, dict), f"expected dict, got: {func}"
    return func


def function_exists(
    base_url: str,
    auth_headers: dict[str, str],
    project_id: str,
    app_id: str,
    func_id: str,
) -> bool:
    # https://services.cloud.mongodb.com/api/admin/v3.0/groups/{groupId}/apps/{appId}/functions/{functionId}
    get_func_response = requests.get(
        f"{base_url}api/admin/v3.0/groups/{project_id}/apps/{app_id}/functions/{func_id}",
        headers=auth_headers,
        timeout=10,
    )
    if get_func_response.status_code == 404:  # noqa: PLR2004
        return False
    get_func_response.raise_for_status()
    func = get_func_response.json()
    assert isinstance(func, dict), f"expected dict response, got: {func}"
    return True


class _RetryPostRequestError(Exception):
    pass


@retry(
    stop=stop_after_attempt(5),
    wait=wait_fixed(10),
    retry=retry_if_exception_type(),
    reraise=True,
)
def _request_post_call(
    url: str,
    data: dict,
    headers: dict[str, str],
    timeout: int,
    *,
    log_data_on_failure: bool = False,
) -> dict:
    response = requests.post(url, json=data, headers=headers, timeout=timeout)
    if response.status_code >= 500:  # noqa: PLR2004
        logger.warning(f"failed to post to {url}, status_code: {response.status_code}, response: {response.text}")
        if log_data_on_failure:
            logger.warning(f"data: {data}")
        raise _RetryPostRequestError(f"status_code: {response.status_code}, response: {response.text}")
    response.raise_for_status()
    return response.json()



from __future__ import annotations

import asyncio
import logging
from collections import Counter
from dataclasses import dataclass, field
from datetime import date, datetime, timedelta
from enum import StrEnum
from functools import reduce, total_ordering
from pathlib import Path
import re
from typing import Callable, ClassVar, TypeVar

from ask_shell.rich_progress import new_task
from model_lib import Entity
from pydantic import Field, model_validator
from zero_3rdparty import datetime_utils, file_utils
from zero_3rdparty.iter_utils import group_by_once

from atlas_init.cli_tf.github_logs import summary_dir
from atlas_init.cli_tf.go_test_run import GoTestRun, GoTestStatus
from atlas_init.cli_tf.go_test_tf_error import (
    GoTestError,
    GoTestErrorClass,
    GoTestErrorClassification,
    details_short_description,
    parse_error_details,
)
from atlas_init.crud.mongo_dao import MongoDao, init_mongo_dao
from atlas_init.html_out.md_export import MonthlyReportPaths
from atlas_init.settings.env_vars import AtlasInitSettings

logger = logging.getLogger(__name__)
_COMPLETE_STATUSES = {GoTestStatus.PASS, GoTestStatus.FAIL}


@total_ordering
class GoTestSummary(Entity):
    name: str
    results: list[GoTestRun] = Field(default_factory=list)
    classifications: dict[str, GoTestErrorClassification] = Field(default_factory=dict)

    @model_validator(mode="after")
    def sort_results(self):
        self.results.sort()
        return self

    @property
    def total_completed(self) -> int:
        return sum((r.status in _COMPLETE_STATUSES for r in self.results), 0)

    @property
    def success_rate(self) -> float:
        total = self.total_completed
        if total == 0:
            if not self.is_skipped:
                logger.warning(f"No results to calculate success rate for {self.name}")
            return 0
        return sum(r.status == "PASS" for r in self.results) / total

    @property
    def is_skipped(self) -> bool:
        return all(r.status == GoTestStatus.SKIP for r in self.results)

    @property
    def success_rate_human(self) -> str:
        return f"{self.success_rate:.2%}"

    @property
    def group_name(self) -> str:
        return next((r.group_name for r in self.results if r.group_name), "unknown-group")

    def last_pass_human(self) -> str:
        return next(
            (f"Passed {test.when}" for test in reversed(self.results) if test.status == GoTestStatus.PASS),
            "never passed",
        )

    def __lt__(self, other) -> bool:
        if not isinstance(other, GoTestSummary):
            raise TypeError
        return (self.success_rate, self.name) < (other.success_rate, other.name)

    def select_tests(self, date: date) -> list[GoTestRun]:
        return [r for r in self.results if r.ts.date() == date]


def summary_str(summary: GoTestSummary, start_date: datetime, end_date: datetime) -> str:
    return "\n".join(
        [
            f"## {summary.name}",
            f"Success rate: {summary.success_rate_human}",
            "",
            "### Timeline",
            *timeline_lines(summary, start_date, end_date),
            "",
            *failure_details(summary),
        ]
    )


def test_detail_md(summary: GoTestSummary, start_date: datetime, end_date: datetime) -> str:
    return "\n".join(
        [
            f"# {summary.name} Test Details",
            summary_line(summary.results),
            f"Success rate: {summary.success_rate_human}",
            "",
            *error_table(summary),
            "## Timeline",
            *timeline_lines(summary, start_date, end_date),
        ]
    )


def timeline_lines(summary: GoTestSummary, start_date: datetime, end_date: datetime) -> list[str]:
    lines = []
    one_day = timedelta(days=1)
    for active_date in datetime_utils.day_range(start_date.date(), (end_date + one_day).date(), one_day):
        active_tests = summary.select_tests(active_date)
        if not active_tests:
            lines.append(f"- {active_date:%Y-%m-%d}: MISSING")
            continue
        lines.append(f"- {active_date:%Y-%m-%d}")
        if len(active_tests) == 1:
            test = active_tests[0]
            if test.is_failure:
                lines.extend(_extract_error_lines(test, summary))
            else:
                lines[-1] += f" {format_test_oneline(test)}"
        if len(active_tests) > 1:
            for test in active_tests:
                error_lines = _extract_error_lines(test, summary)
                lines.extend(
                    [
                        f"  - {format_test_oneline(test)}",
                        *error_lines,
                    ]
                )
    return lines


def _error_header(test: GoTestRun) -> str:
    return f"Error {test.ts.isoformat('T', timespec='seconds')}"


def _extract_error_lines(test: GoTestRun, summary: GoTestSummary) -> list[str]:
    if not test.is_failure:
        return []
    error_classification = summary.classifications.get(test.id)
    classification_lines = [str(error_classification)] if error_classification else []
    details_lines = [details_short_description(error_classification.details)] if error_classification else []
    return [
        "",
        f"### {_error_header(test)}",
        *classification_lines,
        *details_lines,
        f"```\n{test.output_lines_str}\n```",
        "",
    ]


def failure_details(summary: GoTestSummary) -> list[str]:
    lines = ["## Failures"]
    for test in summary.results:
        if test.status == GoTestStatus.FAIL:
            lines.extend(
                (
                    f"### {test.when} {format_test_oneline(test)}",
                    test.finish_summary(),  # type: ignore
                    "",
                )
            )
    return lines


def error_table(summary: GoTestSummary) -> list[str]:
    error_rows: list[dict] = []
    for test in summary.results:
        if test.is_failure:
            anchor = header_to_markdown_link(_error_header(test))
            row = {
                "Date": f"[{test.ts.strftime('%Y-%m-%d %H:%M')}]({anchor})",
                "Env": test.env,
                "Runtime": f"{test.run_seconds:.2f}s",
            }
            error_rows.append(row)
            if error_cls := summary.classifications.get(test.id):
                row["Error Class"] = error_cls.error_class
                row["Details"] = details_short_description(error_cls.details)
            elif auto_class := GoTestErrorClass.auto_classification(test.output_lines_str):
                row["Error Class"] = auto_class
            if "Details" not in row:
                row["Details"] = details_short_description(parse_error_details(test))
    if not error_rows:
        return []
    headers = sorted(reduce(lambda x, y: x.union(y.keys()), error_rows, set()))
    return markdown_table_lines("Error Table", error_rows, headers, lambda row: [row.get(key, "") for key in headers])


def format_test_oneline(test: GoTestRun) -> str:
    if job_url := test.job_url:
        return f"[{test.status} {test.runtime_human}]({job_url})"
    return f"{test.status} {test.runtime_human}"  # type: ignore


def header_to_markdown_link(header: str) -> str:
    """
    Converts a markdown header to a markdown link anchor.
    Example:
        'Error 2025-05-23T00:28:50+00:00' -> '#error-2025-05-23t0028500000'
    """
    anchor = header.strip().lower()
    # Remove all characters except alphanumerics, spaces, and hyphens
    anchor = re.sub(r"[^a-z0-9 \-]", "", anchor)
    anchor = anchor.replace(" ", "-")
    return f"#{anchor}"


def create_detailed_summary(
    summary_name: str,
    end_test_date: datetime,
    start_test_date: datetime,
    test_results: dict[str, list[GoTestRun]],
    expected_names: set[str] | None = None,
) -> list[str]:
    summary_dir_path = summary_dir(summary_name)
    if summary_dir_path.exists():
        file_utils.clean_dir(summary_dir_path)
    summaries = [GoTestSummary(name=name, results=runs) for name, runs in test_results.items()]
    top_level_summary = ["# SUMMARY OF ALL TESTS name (success rate)"]
    summaries = [summary for summary in summaries if summary.results and not summary.is_skipped]
    if expected_names and (skipped_names := expected_names - {summary.name for summary in summaries}):
        logger.warning(f"skipped test names: {'\n'.join(skipped_names)}")
        top_level_summary.append(f"Skipped tests: {', '.join(skipped_names)}")
    for summary in sorted(summaries):
        test_summary_path = summary_dir_path / f"{summary.success_rate_human}_{summary.name}.md"
        test_summary_md = summary_str(summary, start_test_date, end_test_date)
        file_utils.ensure_parents_write_text(test_summary_path, test_summary_md)
        top_level_summary.append(
            f"- {summary.name} - {summary.group_name} ({summary.success_rate_human}) ({summary.last_pass_human()}) ('{test_summary_path}')"
        )
    return top_level_summary


def create_short_summary(test_results: dict[str, list[GoTestRun]], failing_names: list[str]) -> list[str]:
    summary = ["# SUMMARY OF FAILING TESTS"]
    summary_fail_details: list[str] = ["# FAIL DETAILS"]

    for fail_name in failing_names:
        fail_tests = test_results[fail_name]
        summary.append(f"- {fail_name} has {len(fail_tests)} failures:")
        summary.extend(
            f"  - [{fail_run.when} failed in {fail_run.runtime_human}]({fail_run.url})"  # type: ignore
            for fail_run in fail_tests
        )
        summary_fail_details.append(f"\n\n ## {fail_name} details:")
        summary_fail_details.extend(f"```\n{fail_run.finish_summary()}\n```" for fail_run in fail_tests)  # type: ignore
    logger.info("\n".join(summary_fail_details))
    return summary


@dataclass
class GoRunTestReport:
    summary: str
    error_details: str


def create_test_report(
    runs: list[GoTestRun],
    errors: list[GoTestError],
    *,
    indent_size=2,
    max_runs=20,
    env_name: str = "",
) -> GoRunTestReport:
    if env_name:
        runs = [run for run in runs if run.env == env_name]
        errors = [error for error in errors if error.run.env == env_name]
    single_indent = " " * indent_size
    if not runs:
        return GoRunTestReport(
            summary="No test runs found",
            error_details="",
        )
    envs = {run.env for run in runs if run.env}
    lines = [summary_line(runs)]
    if errors:
        env_name_str = f" in {env_name}" if env_name else ""
        lines.append(f"\n\n## Errors Overview{env_name_str}")
        lines.extend(error_overview_lines(errors, single_indent))
    for env in envs:
        env_runs = [run for run in runs if run.env == env]
        lines.append(f"\n\n## {env.upper()} Had {len(env_runs)} Runs")
        lines.extend(env_summary_lines(env_runs, max_runs, single_indent))
    if len(envs) > 1:
        lines.append(f"\n\n## All Environments Had {len(runs)} Runs")
        lines.extend(env_summary_lines(runs, max_runs, single_indent))
    error_detail_lines = []
    if errors:
        error_detail_lines.append("# Errors Details")
        error_detail_lines.extend(error_details(errors, include_env=len(envs) > 1))
    return GoRunTestReport(
        summary="\n".join(lines),
        error_details="\n".join(error_detail_lines),
    )


def run_statuses(runs: list[GoTestRun]) -> str:
    if counter := Counter([run.status for run in runs]):
        return " ".join(
            f"{cls}(x {count})" if count > 1 else cls
            for cls, count in sorted(counter.items(), key=lambda item: item[1], reverse=True)
        )
    return ""


def summary_line(runs: list[GoTestRun]):
    run_delta = GoTestRun.run_delta(runs)
    envs = {run.env for run in runs if run.env}
    pkg_test_names = {run.name_with_package for run in runs}
    envs_str = ", ".join(sorted(envs))
    branches = {run.branch for run in runs if run.branch}
    branches_str = (
        "from " + ", ".join(sorted(branches)) + " branches" if len(branches) > 1 else f"from {branches.pop()} branch"
    )
    return f"# Found {len(runs)} TestRuns in {envs_str} {run_delta} {branches_str}: {len(pkg_test_names)} unique tests, {run_statuses(runs)}"


def error_overview_lines(errors: list[GoTestError], single_indent: str) -> list[str]:
    lines = []
    grouped_errors = GoTestError.group_by_classification(errors)
    if errors_unclassified := grouped_errors.unclassified:
        lines.append(f"- Found {len(grouped_errors.unclassified)} unclassified errors:")
        lines.extend(count_errors_by_test(single_indent, errors_unclassified))
    if errors_by_class := grouped_errors.classified:
        for classification, errors in errors_by_class.items():
            lines.append(f"- Error Type `{classification}`:")
            lines.extend(count_errors_by_test(single_indent, errors))
    return lines


def count_errors_by_test(indent: str, errors: list[GoTestError]) -> list[str]:
    lines: list[str] = []
    counter = Counter()
    for error in errors:
        counter[error.header(use_ticks=True)] += 1
    for error_header, count in counter.most_common():
        if count > 1:
            lines.append(f"{indent}- {count} x {error_header}")
        else:
            lines.append(f"{indent}- {error_header}")
    return sorted(lines)


def env_summary_lines(env_runs: list[GoTestRun], max_runs: int, single_indent: str) -> list[str]:
    lines: list[str] = []
    if pass_rates := GoTestRun.lowest_pass_rate(env_runs, max_tests=max_runs, include_single_run=False):
        lines.append(f"- Lowest pass rate: {GoTestRun.run_delta(env_runs)}")
        for pass_rate, name, name_tests in pass_rates:
            ran_count_str = f"ran {len(name_tests)} times" if len(name_tests) > 1 else "ran 1 time"
            if last_pass := GoTestRun.last_pass(name_tests):
                lines.append(f"{single_indent}- {pass_rate:.2%} {name} ({ran_count_str}) last PASS {last_pass}")
            else:
                lines.append(f"{single_indent}- {pass_rate:.2%} {name} ({ran_count_str}) never passed")
    if pass_stats := GoTestRun.last_pass_stats(env_runs, max_tests=max_runs):
        lines.append(f"- Longest time since `{GoTestStatus.PASS}`: {GoTestRun.run_delta(env_runs)}")
        lines.extend(
            f"{single_indent}- {pass_stat.pass_when} {pass_stat.name_with_package}" for pass_stat in pass_stats
        )
    lines.append(f"- Slowest tests: {GoTestRun.run_delta(env_runs)}")
    for time_stat in GoTestRun.slowest_tests(env_runs):
        avg_time_str = (
            f"(avg = {time_stat.average_duration} across {len(time_stat.runs)} runs)"
            if time_stat.average_seconds
            else ""
        )
        lines.append(
            f"{single_indent}- {time_stat.slowest_duration} {time_stat.name_with_package} {avg_time_str}".rstrip()
        )
    return lines


def error_details(errors: list[GoTestError], include_env: bool) -> list[str]:
    lines: list[str] = []
    for name, name_errors in GoTestError.group_by_name_with_package(errors).items():
        lines.append(
            f"## {name} had {len(name_errors)} errors {GoTestRun.run_delta([error.run for error in name_errors])}",
        )
        for error in sorted(name_errors, reverse=True):  # newest first
            env_str = f" in {error.run.env} " if include_env and error.run.env else ""
            lines.extend(
                [
                    f"### Started @ {error.run.ts} {env_str}ran for ({error.run.runtime_human})",
                    f"- error classes: bot={error.bot_error_class}, human={error.human_error_class}",
                    f"- details summary: {error.short_description}",
                    f"- test output:\n```log\n{error.run.output_lines_str}\n```\n",
                ]
            )
    return lines


class TFCITestOutput(Entity):
    """Represent the CI Test Output for a day"""

    log_paths: list[Path] = Field(
        default_factory=list, description="Paths to the log files of the test runs analyzed by the run history."
    )
    found_tests: list[GoTestRun] = Field(default_factory=list, description="All tests for report day.")
    found_errors: list[GoTestError] = Field(default_factory=list, description="All errors for the report day.")
    classified_errors: list[GoTestErrorClassification] = Field(
        default_factory=list, description="Classified errors for the report day."
    )


class RunHistoryFilter(Entity):
    run_history_start: datetime
    run_history_end: datetime
    env_filter: list[str] = Field(default_factory=list)
    skip_branch_filter: bool = False


class MonthlyReportIn(Entity):
    name: str
    branch: str
    history_filter: RunHistoryFilter
    skip_columns: set[ErrorRowColumns] = Field(default_factory=set)
    skip_rows: list[Callable[[TestRow], bool]] = Field(default_factory=list)
    existing_details_md: dict[str, str] = Field(default_factory=dict)
    report_paths: MonthlyReportPaths

    @classmethod
    def skip_skipped(cls, test: TestRow) -> bool:
        return all(run.is_skipped for runs in test.last_env_runs.values() for run in runs)

    @classmethod
    def skip_if_no_failures(cls, test: TestRow) -> bool:
        return not any(run.is_failure for runs in test.last_env_runs.values() for run in runs)


class MonthlyReportOut(Entity):
    summary_md: str
    test_details_md: dict[str, str] = Field(default_factory=dict)


class ErrorRowColumns(StrEnum):
    GROUP_NAME = "Group with Package"
    TEST = "Test"
    ERROR_CLASS = "Error Class"
    DETAILS_SUMMARY = "Details Summary"
    PASS_RATE = "Pass Rate"  # nosec B105 # This is not a security issue, just a column name
    TIME_SINCE_PASS = "Time Since PASS"  # nosec B105 # This is not a security issue, just a column name

    __ENV_BASED__: ClassVar[list[str]] = [PASS_RATE, TIME_SINCE_PASS]

    @classmethod
    def column_names(cls, rows: list[TestRow], skip_columns: set[ErrorRowColumns]) -> list[str]:
        if not rows:
            return []
        envs = set()
        for row in rows:
            envs.update(row.last_env_runs.keys())
        columns: list[str] = [cls.GROUP_NAME, cls.TEST, cls.ERROR_CLASS, cls.DETAILS_SUMMARY]
        for env in sorted(envs):
            columns.extend(f"{env_col} ({env})" for env_col in cls.__ENV_BASED__ if env_col not in skip_columns)
        return [col for col in columns if col not in skip_columns]


@total_ordering
class TestRow(Entity):
    group_name: str
    package_url: str
    full_name: str
    test_name: str
    error_classes: list[GoTestErrorClass]
    details_summary: str
    last_env_runs: dict[str, list[GoTestRun]] = field(default_factory=dict)

    def __lt__(self, other) -> bool:
        if not isinstance(other, TestRow):
            raise TypeError
        return (self.group_name, self.test_name) < (other.group_name, other.test_name)

    @property
    def pass_rates(self) -> dict[str, float]:
        rates = {}
        for env, runs in self.last_env_runs.items():
            if not runs:
                continue
            total = len(runs)
            passed = sum(run.status == GoTestStatus.PASS for run in runs)
            rates[env] = passed / total if total > 0 else 0.0
        return rates

    @property
    def time_since_pass(self) -> dict[str, str]:
        time_since = {}
        for env, runs in self.last_env_runs.items():
            if not runs:
                time_since[env] = "never run"
                continue
            time_since[env] = next(
                (run.ts.strftime("%Y-%m-%d") for run in sorted(runs, reverse=True) if run.status == GoTestStatus.PASS),
                "never pass",
            )
        return time_since

    @property
    def error_classes_str(self) -> str:
        if counter := Counter(self.error_classes):
            return " ".join(
                f"{cls}(x {count})" if count > 1 else cls
                for cls, count in sorted(counter.items(), key=lambda item: item[1], reverse=True)
            )
        return "No error classes"

    def as_row(self, columns: list[str]) -> list[str]:
        values = []
        pass_rates = self.pass_rates
        time_since_pass = self.time_since_pass
        for col in columns:
            match col:
                case ErrorRowColumns.GROUP_NAME:
                    group_part = self.full_name.removesuffix(self.test_name).rstrip("/")
                    values.append(group_part or "Unknown Group")
                case ErrorRowColumns.TEST:
                    values.append(self.test_name)
                case ErrorRowColumns.ERROR_CLASS:
                    values.append(self.error_classes_str)
                case ErrorRowColumns.DETAILS_SUMMARY:
                    values.append(self.details_summary)
                case s if s.startswith(ErrorRowColumns.PASS_RATE):
                    env = s.split(" (")[-1].rstrip(")")
                    env_pass_rate = pass_rates.get(env, 0.0)
                    env_run_count = len(self.last_env_runs.get(env, []))
                    pass_rate_pct = f"{env_pass_rate:.2%} ({env_run_count} runs)" if env in pass_rates else "N/A"
                    if pass_rate_pct.startswith("100.00%"):
                        values.append("always")  # use always to avoid sorting errors, 100% showing before 2%
                    else:
                        values.append(pass_rate_pct)
                case s if s.startswith(ErrorRowColumns.TIME_SINCE_PASS):
                    env = s.split(" (")[-1].rstrip(")")
                    values.append(time_since_pass.get(env, "never passed"))
                case _:
                    logger.warning(f"Unknown column: {col}, skipping")
                    values.append("N/A")
        return values


def create_monthly_report(settings: AtlasInitSettings, event: MonthlyReportIn) -> MonthlyReportOut:
    with new_task(f"Monthly Report for {event.name} on {event.branch}"):
        test_rows, detail_files_md = asyncio.run(_collect_monthly_test_rows_and_summaries(settings, event))
        assert test_rows, "No error rows found for monthly report"
    columns = ErrorRowColumns.column_names(test_rows, event.skip_columns)
    skip_rows = (
        []
        if event.skip_rows == []
        else [
            "",
            "## Skip Test Filters",
            *[f"- {method.__name__}" for method in event.skip_rows],
            "",
        ]
    )
    summary_md = [
        f"# Monthly Report for {event.name} on {event.branch} from {event.history_filter.run_history_start:%Y-%m-%d} to {event.history_filter.run_history_end:%Y-%m-%d} Found {len(test_rows)} unique Tests",
        *skip_rows,
        *markdown_table_lines("Test Run Table", test_rows, columns, lambda row: row.as_row(columns)),
    ]
    return MonthlyReportOut(
        summary_md="\n".join(summary_md),
        test_details_md=detail_files_md,
    )


class DailyReportIn(Entity):
    report_date: datetime
    history_filter: RunHistoryFilter
    skip_columns: set[ErrorRowColumns] = Field(default_factory=set)
    row_modifier: Callable[[TestRow, dict[str, str]], dict[str, str]] | None = Field(
        default=None, description=f"Use the {ErrorRowColumns} to access column-->value mapping"
    )


class DailyReportOut(Entity):
    summary_md: str
    details_md: str


T = TypeVar("T")


def markdown_table_lines(
    header: str, rows: list[T], columns: list[str], row_to_line: Callable[[T], list[str]], *, header_level: int = 2
) -> list[str]:
    if not rows:
        return []
    return [
        f"{'#' * header_level} {header}",
        "",
        " | ".join(columns),
        " | ".join("---" for _ in columns),
        *(" | ".join(row_to_line(row)) for row in rows),
        "",
    ]


def create_daily_report(output: TFCITestOutput, settings: AtlasInitSettings, event: DailyReportIn) -> DailyReportOut:
    errors = output.found_errors
    error_classes = {cls.run_id: cls.error_class for cls in output.classified_errors}
    one_line_summary = summary_line(output.found_tests)

    with new_task("Daily Report"):
        with new_task("Collecting error rows") as task:
            failure_rows = asyncio.run(
                _collect_daily_error_rows(errors, error_classes, settings, event.history_filter, task)
            )
        if not failure_rows:
            return DailyReportOut(summary_md=f"🎉All tests passed\n{one_line_summary}", details_md="")
    columns = ErrorRowColumns.column_names(failure_rows, event.skip_columns)

    def as_md_row(row: TestRow) -> list[str]:
        if row_modifier := event.row_modifier:
            row_dict = dict(zip(columns, row.as_row(columns)))
            row_dict = row_modifier(row, row_dict)
            return [row_dict[col] for col in columns]
        return row.as_row(columns)

    summary_md = [
        f"# Daily Report on {event.report_date:%Y-%m-%d}",
        one_line_summary,
        "",
        *markdown_table_lines("Errors Table", failure_rows, columns, as_md_row),
    ]
    return DailyReportOut(summary_md="\n".join(summary_md), details_md="TODO")


async def _collect_daily_error_rows(
    errors: list[GoTestError],
    error_classes: dict[str, GoTestErrorClass],
    settings: AtlasInitSettings,
    event: RunHistoryFilter,
    task: new_task,
) -> list[TestRow]:
    error_rows: list[TestRow] = []
    dao = await init_mongo_dao(settings)
    for error in errors:
        test_run = error.run
        error_class = error_classes[error.run_id]
        summary = error.short_description
        error_row, _ = await _create_test_row(event, dao, test_run, error_class, summary)
        error_rows.append(error_row)
        task.update(advance=1)
    return sorted(error_rows)


async def _collect_monthly_test_rows_and_summaries(
    settings: AtlasInitSettings,
    event: MonthlyReportIn,
) -> tuple[list[TestRow], dict[str, str]]:
    dao = await init_mongo_dao(settings)
    branch = event.branch
    history_filter = event.history_filter
    summary_name = event.name
    skip_rows = event.skip_rows
    last_day_test_names = await dao.read_tf_tests_for_day(branch, history_filter.run_history_end)
    test_runs_by_name: dict[str, GoTestRun] = {run.full_name: run for run in last_day_test_names}
    test_rows = []
    detail_files_md: dict[str, str] = {}
    with new_task("Collecting monthly error rows", total=len(last_day_test_names)) as task:
        for name_with_group, test_run in test_runs_by_name.items():
            test_row, runs = await _create_test_row(
                history_filter,
                dao,
                test_run,
            )
            if any(skip(test_row) for skip in skip_rows):
                continue
            test_rows.append(test_row)
            run_ids = [run.id for run in runs]
            classifications = await dao.read_error_classifications(run_ids)
            test_row.error_classes = [cls.error_class for cls in classifications.values()]
            test_row.details_summary = (
                f"[{run_statuses(runs)}]({settings.github_ci_summary_details_rel_path(summary_name, name_with_group)})"
            )
            if name_with_group not in event.existing_details_md:
                summary = GoTestSummary(name=name_with_group, results=runs, classifications=classifications)
                detail_files_md[name_with_group] = test_detail_md(
                    summary, history_filter.run_history_start, history_filter.run_history_end
                )
            task.update(advance=1)
    return sorted(test_rows), detail_files_md


async def _create_test_row(
    history_filter: RunHistoryFilter,
    dao: MongoDao,
    test_run: GoTestRun,
    error_class: GoTestErrorClass | None = None,
    summary: str = "",
) -> tuple[TestRow, list[GoTestRun]]:
    package_url = test_run.package_url
    group_name = test_run.group_name
    package_url = test_run.package_url or ""
    branch = test_run.branch
    branch_filter = []
    test_name = test_run.name
    if branch and not history_filter.skip_branch_filter:
        branch_filter.append(branch)
    run_history = await dao.read_run_history(
        test_name=test_name,
        package_url=package_url,
        group_name=group_name,
        start_date=history_filter.run_history_start,
        end_date=history_filter.run_history_end,
        envs=history_filter.env_filter,
        branches=branch_filter,
    )
    last_env_runs = group_by_once(run_history, key=lambda run: run.env or "unknown-env")
    error_classes = [error_class] if error_class else []
    return TestRow(
        full_name=test_run.full_name,
        group_name=group_name,
        package_url=package_url,
        test_name=test_name,
        error_classes=error_classes,
        details_summary=summary,
        last_env_runs=last_env_runs,
    ), run_history



import logging
from functools import singledispatch
from queue import Queue

from atlas_init.cli_tf.schema_v2 import (
    extend_import_urls,
    go_fmt,
    import_lines,
    package_name,
)
from atlas_init.cli_tf.schema_v2_sdk import GoVarName, SDKAttribute, SDKModel
from atlas_init.cli_tf.schema_v3 import (
    TF_MODEL_NAME,
    Attribute,
    ListNestedAttribute,
    Resource,
    SingleNestedAttribute,
)
from atlas_init.cli_tf.schema_v3_sdk_base import (
    AllowedMissingAttributeError,
    SDKAndSchemaAttribute,
    find_attribute,
    name_schema_struct,
    name_struct_attribute,
    schema_attributes,
)
from atlas_init.humps import camelize, pascalize

logger = logging.getLogger(__name__)


def generate_model_go(resource: Resource, sdk_model: SDKModel) -> str:
    func_lines = sdk_to_tf_func(resource, sdk_model)
    import_urls = set()
    extend_import_urls(import_urls, func_lines)
    unformatted = "\n".join(
        [
            f"package {package_name(resource.name)}",
            "",
            *import_lines(import_urls),
            "",
            *func_lines,
        ]
    )
    return go_fmt(resource.name, unformatted)


def sdk_to_tf_func(resource: Resource, sdk_model: SDKModel) -> list[str]:
    lines = []
    timeouts_signature = ", timeout timeouts.Value" if resource.use_timeout else ""
    lines.append(
        f"func New{TF_MODEL_NAME}(ctx context.Context, {GoVarName.INPUT} *admin.{sdk_model.name}{timeouts_signature}) (*{TF_MODEL_NAME}, diag.Diagnostics) {{"
    )
    nested_attributes, call_lines = call_nested_functions(resource, sdk_model.list_nested_attributes())
    lines.extend(call_lines)
    timeouts_set = ["    Timeouts: timeout,"] if resource.use_timeout else []
    lines.extend(
        [
            f"  return &{TF_MODEL_NAME}{{",
            *tf_struct_create(resource, sdk_model),
            *timeouts_set,
            "  }, nil",  # close return
            "}\n",  # close function
        ]
    )
    lines.extend(process_nested_attributes(nested_attributes))
    return lines


def as_var_name(attr: SDKAttribute) -> str:
    return camelize(attr.json_name)


def call_nested_functions(
    root: Resource | Attribute,
    nested_attributes: list[SDKAttribute],
    *,
    return_on_error: bool = True,
    sdk_var_name: GoVarName = GoVarName.INPUT,
) -> tuple[list[SDKAndSchemaAttribute], list[str]]:
    lines = []
    schema_nested_attributes = schema_attributes(root)
    nested_generations = []
    for sdk_attribute in nested_attributes:
        try:
            schema_attribute = find_attribute(schema_nested_attributes, sdk_attribute.tf_name, root.name)
        except AllowedMissingAttributeError as e:
            logger.info(f"skipping {e!r}")
            continue
        var_name = as_var_name(sdk_attribute)
        lines.append(
            f"{var_name} := New{_name_custom_object_type(schema_attribute.name)}(ctx, {sdk_var_name}.{sdk_attribute.struct_name}, {GoVarName.DIAGS})"
        )
        nested_generations.append(SDKAndSchemaAttribute(sdk_attribute, schema_attribute))
    if lines and return_on_error:
        lines.insert(0, "diags := &diag.Diagnostics{}")
        lines.extend(
            [
                "  if diags.HasError() {",
                "    return nil, *diags",
                "  }",
            ]
        )

    return nested_generations, lines


_sdk_to_tf_funcs = {
    ("*string", "string"): lambda sdk_ref: f"types.StringPointerValue({sdk_ref})",
    ("string", "string"): lambda sdk_ref: f"types.StringValue({sdk_ref})",
    ("*int64", "int64"): lambda sdk_ref: f"types.Int64PointerValue({sdk_ref})",
    ("int64", "int64"): lambda sdk_ref: f"types.Int64Value({sdk_ref})",
    (
        "*int",
        "int64",
    ): lambda sdk_ref: f"types.Int64PointerValue(conversion.IntPtrToInt64Ptr({sdk_ref}))",
    ("*float64", "float64"): lambda sdk_ref: f"types.Float64PointerValue({sdk_ref})",
    ("float64", "float64"): lambda sdk_ref: f"types.Float64Value({sdk_ref})",
    ("*bool", "bool"): lambda sdk_ref: f"types.BoolPointerValue({sdk_ref})",
    ("bool", "bool"): lambda sdk_ref: f"types.BoolValue({sdk_ref})",
    (
        "*map[string]string",
        "map[string]string",
    ): lambda sdk_ref: f"conversion.ToTFMapOfString({GoVarName.CTX}, {GoVarName.DIAGS}, {sdk_ref})",
    (
        "map[string]string",
        "map[string]string",
    ): lambda sdk_ref: f"conversion.ToTFMapOfString({GoVarName.CTX}, {GoVarName.DIAGS}, &{sdk_ref})",
    (
        "*time.Time",
        "string",
    ): lambda sdk_ref: f"types.StringPointerValue(conversion.TimePtrToStringPtr({sdk_ref}))",
    (
        "time.Time",
        "string",
    ): lambda sdk_ref: f"types.StringValue(conversion.TimeToString({sdk_ref}))",
}


def sdk_to_tf_attribute_value(
    schema_attribute: Attribute,
    sdk_attribute: SDKAttribute,
    variable_name: GoVarName = GoVarName.INPUT,
) -> str:
    key = (sdk_attribute.go_type, schema_attribute.go_type)
    if key in _sdk_to_tf_funcs:
        return _sdk_to_tf_funcs[key](f"{variable_name}.{sdk_attribute.struct_name}")
    raise ValueError(f"Could not find conversion function for {key}")


# sdk_to_tf_attribute_value(schema_attribute, sdk_attribute, sdk_var_name)
def tf_struct_create(
    root: Resource | Attribute,
    sdk_model: SDKModel,
    sdk_var_name: GoVarName = GoVarName.INPUT,
) -> list[str]:
    lines = []
    for attr in schema_attributes(root):
        if attr.is_nested:
            local_var = sdk_model.lookup_tf_name(attr.name)
            lines.append(f"{name_struct_attribute(attr.name)}: {as_var_name(local_var)},")
        elif attr.is_attribute:
            local_var = sdk_model.lookup_tf_name(attr.name)
            lines.append(
                f"{name_struct_attribute(attr.name)}: {sdk_to_tf_attribute_value(attr, local_var, sdk_var_name)},"
            )
    return lines


def process_nested_attributes(
    nested_attributes: list[SDKAndSchemaAttribute],
) -> list[str]:
    lines = []
    queue = Queue()

    def add_nested_to_queue(attributes: list[SDKAndSchemaAttribute]):
        for nested in attributes:
            logger.info(f"found nested attribute: {nested.schema_attribute.name}")
            queue.put(nested)

    add_nested_to_queue(nested_attributes)
    while not queue.empty():
        sdk_attribute, schema_attribute = queue.get()
        more_nested_attributes, nested_lines = convert_nested_attribute(
            schema_attribute.nested_model, schema_attribute, sdk_attribute
        )
        lines.extend(nested_lines)
        add_nested_to_queue(more_nested_attributes)
    return lines


def _name_custom_object_type(name: str) -> str:
    return f"{pascalize(name)}ObjType"


@singledispatch
def convert_nested_attribute(
    nested_model: object, schema_attribute: Attribute, _: SDKAttribute
) -> tuple[list[SDKAndSchemaAttribute], list[str]]:
    raise NotImplementedError(f"unsupported nested attribute: {schema_attribute.name} of type {type(nested_model)}")


@convert_nested_attribute.register
def _convert_single_nested_attribute(
    _: SingleNestedAttribute,
    schema_attribute: Attribute,
    sdk_attribute: SDKAttribute,
) -> tuple[list[SDKAndSchemaAttribute], list[str]]:
    object_type_name = _name_custom_object_type(schema_attribute.name)
    lines: list[str] = [
        f"func New{object_type_name}(ctx context.Context, {GoVarName.INPUT} *admin.{sdk_attribute.struct_type_name}, diags *diag.Diagnostics) types.Object {{",
        f"  if {GoVarName.INPUT} == nil {{",
        f"    return types.ObjectNull({object_type_name}.AttrTypes)",
        "  }",
    ]

    nested_attributes, call_lines = call_nested_functions(
        schema_attribute, sdk_attribute.list_nested_attributes(), return_on_error=False
    )
    lines.extend(call_lines)
    struct_name = name_schema_struct(schema_attribute.name)
    lines.extend(
        [
            f"  tfModel := {struct_name}{{",
            *tf_struct_create(schema_attribute, sdk_attribute.as_sdk_model()),
            "  }",
            f"  objType, diagsLocal := types.ObjectValueFrom(ctx, {object_type_name}.AttrTypes, tfModel)",
            f"  {GoVarName.DIAGS}.Append(diagsLocal...)",
            "  return objType",
            "}\n",
        ]
    )
    return nested_attributes, lines


@convert_nested_attribute.register
def _convert_list_nested_attriute(
    _: ListNestedAttribute,
    schema_attribute: Attribute,
    sdk_attribute: SDKAttribute,
) -> tuple[list[SDKAndSchemaAttribute], list[str]]:
    object_type_name = _name_custom_object_type(schema_attribute.name)
    lines: list[str] = [
        f"func New{object_type_name}(ctx context.Context, {GoVarName.INPUT} *[]admin.{sdk_attribute.struct_type_name}, diags *diag.Diagnostics) types.List {{",
        f"  if {GoVarName.INPUT} == nil {{",
        f"    return types.ListNull({object_type_name})",
        "  }",
    ]
    struct_name = name_schema_struct(schema_attribute.name)
    lines.extend(
        [
            f"  tfModels := make([]{struct_name}, len(*{GoVarName.INPUT}))",
            f"  for i, item := range *{GoVarName.INPUT} {{",
        ]
    )
    nested_attributes, call_lines = call_nested_functions(
        schema_attribute,
        sdk_attribute.list_nested_attributes(),
        return_on_error=False,
        sdk_var_name=GoVarName.ITEM,
    )
    lines.extend([f"  {line}" for line in call_lines])
    lines.extend(
        [
            f"    tfModels[i] = {struct_name}{{",
            *tf_struct_create(
                schema_attribute,
                sdk_attribute.as_sdk_model(),
                sdk_var_name=GoVarName.ITEM,
            ),
            "    }",
            "  }",
            f"  listType, diagsLocal := types.ListValueFrom(ctx, {object_type_name}, tfModels)",
            f"  {GoVarName.DIAGS}.Append(diagsLocal...)",
            "  return listType",
            "}\n",
        ]
    )
    return nested_attributes, lines



_dry_run = False


def is_dry_run():
    return _dry_run


def set_dry_run(dry_run: bool):
    global _dry_run  # noqa: PLW0603 # instead of passing dry_run everywhere we can use this global variable
    _dry_run = dry_run
